import json
import uuid
from typing import Any, List, Optional

from deepeval.model_integrations.types import InputParameters, OutputParameters
from deepeval.test_case.llm_test_case import ToolCall
from deepeval.tracing.context import (
    current_span_context,
    current_trace_context,
    update_current_span,
    update_llm_span,
)
from deepeval.tracing.trace_context import current_llm_context
from deepeval.tracing.types import ToolSpan, TraceSpanStatus
from deepeval.utils import shorten, len_long


def _update_all_attributes(
    input_parameters: InputParameters,
    output_parameters: OutputParameters,
    expected_tools: List[ToolCall],
    expected_output: str,
    context: List[str],
    retrieval_context: List[str],
):
    """Update span and trace attributes with input/output parameters."""
    update_current_span(
        input=input_parameters.input or input_parameters.messages or "NA",
        output=output_parameters.output or "NA",
        tools_called=output_parameters.tools_called,
        # attributes to be added
        expected_output=expected_output,
        expected_tools=expected_tools,
        context=context,
        retrieval_context=retrieval_context,
    )

    llm_context = current_llm_context.get()

    update_llm_span(
        input_token_count=output_parameters.prompt_tokens,
        output_token_count=output_parameters.completion_tokens,
        prompt=llm_context.prompt,
    )

    if output_parameters.tools_called:
        create_child_tool_spans(output_parameters)

    __update_input_and_output_of_current_trace(
        input_parameters, output_parameters
    )


def __update_input_and_output_of_current_trace(
    input_parameters: InputParameters, output_parameters: OutputParameters
):

    current_trace = current_trace_context.get()
    if current_trace:
        if current_trace.input is None:
            current_trace.input = (
                input_parameters.input or input_parameters.messages
            )
        if current_trace.output is None:
            current_trace.output = output_parameters.output

    return


def create_child_tool_spans(output_parameters: OutputParameters):
    if output_parameters.tools_called is None:
        return

    current_span = current_span_context.get()
    for tool_called in output_parameters.tools_called:
        tool_span = ToolSpan(
            **{
                "uuid": str(uuid.uuid4()),
                "trace_uuid": current_span.trace_uuid,
                "parent_uuid": current_span.uuid,
                "start_time": current_span.start_time,
                "end_time": current_span.start_time,
                "status": TraceSpanStatus.SUCCESS,
                "children": [],
                "name": tool_called.name,
                "input": tool_called.input_parameters,
                "output": None,
                "metrics": None,
                "description": tool_called.description,
            }
        )
        current_span.children.append(tool_span)


_URL_MAX = 200
_JSON_MAX = max(
    len_long(), 400
)  # <- make this bigger by increasing DEEPEVAL_MAXLEN_LONG above 400


def compact_dump(value: Any) -> str:
    try:
        dumped = json.dumps(
            value, ensure_ascii=False, default=str, separators=(",", ":")
        )
    except Exception:
        dumped = repr(value)
    return shorten(dumped, max_len=_JSON_MAX)


def fmt_url(url: Optional[str]) -> str:
    if not url:
        return ""
    if url.startswith("data:"):
        return "[data-uri]"
    return shorten(url, max_len=_URL_MAX)



from deepeval.models.base_model import (
    DeepEvalBaseModel,
    DeepEvalBaseLLM,
    DeepEvalBaseEmbeddingModel,
)
from deepeval.models.llms import (
    GPTModel,
    AzureOpenAIModel,
    LocalModel,
    OllamaModel,
    AnthropicModel,
    GeminiModel,
    AmazonBedrockModel,
    LiteLLMModel,
    KimiModel,
    GrokModel,
    DeepSeekModel,
)
from deepeval.models.embedding_models import (
    OpenAIEmbeddingModel,
    AzureOpenAIEmbeddingModel,
    LocalEmbeddingModel,
    OllamaEmbeddingModel,
)

__all__ = [
    "DeepEvalBaseModel",
    "DeepEvalBaseLLM",
    "DeepEvalBaseEmbeddingModel",
    "GPTModel",
    "AzureOpenAIModel",
    "LocalModel",
    "OllamaModel",
    "AnthropicModel",
    "GeminiModel",
    "AmazonBedrockModel",
    "LiteLLMModel",
    "KimiModel",
    "GrokModel",
    "DeepSeekModel",
    "OpenAIEmbeddingModel",
    "AzureOpenAIEmbeddingModel",
    "LocalEmbeddingModel",
    "OllamaEmbeddingModel",
]



from enum import Enum
import os
import json
from pydantic import BaseModel, Field
from typing import Any, Optional, List, Dict, Union, Tuple
import shutil
import sys
import datetime
from rich.table import Table
from rich.console import Console
from rich import print


from deepeval.metrics import BaseMetric
from deepeval.confident.api import Api, Endpoints, HttpMethods, is_confident
from deepeval.test_run.api import (
    LLMApiTestCase,
    ConversationalApiTestCase,
    TestRunHttpResponse,
    MetricData,
)
from deepeval.tracing.utils import make_json_serializable
from deepeval.tracing.api import SpanApiType, span_api_type_literals
from deepeval.test_case import LLMTestCase, ConversationalTestCase
from deepeval.utils import (
    delete_file_if_exists,
    get_is_running_deepeval,
    is_read_only_env,
    open_browser,
    shorten,
    format_turn,
    len_short,
)
from deepeval.test_run.cache import global_test_run_cache_manager
from deepeval.constants import CONFIDENT_TEST_CASE_BATCH_SIZE, HIDDEN_DIR
from deepeval.prompt import (
    PromptMessage,
    ModelSettings,
    PromptInterpolationType,
    OutputType,
)
from rich.panel import Panel
from rich.columns import Columns


portalocker = None
if not is_read_only_env():
    try:
        import portalocker
    except Exception as e:
        print(
            f"Warning: failed to import portalocker: {e}",
            file=sys.stderr,
        )
else:
    print(
        "Warning: DeepEval is configured for read only environment. Test runs will not be written to disk."
    )


TEMP_FILE_PATH = f"{HIDDEN_DIR}/.temp_test_run_data.json"
LATEST_TEST_RUN_FILE_PATH = f"{HIDDEN_DIR}/.latest_test_run.json"
LATEST_TEST_RUN_DATA_KEY = "testRunData"
LATEST_TEST_RUN_LINK_KEY = "testRunLink"
console = Console()


class TestRunResultDisplay(Enum):
    ALL = "all"
    FAILING = "failing"
    PASSING = "passing"


class MetricScoreType(BaseModel):
    metric: str
    score: float

    @classmethod
    def from_metric(cls, metric: BaseMetric):
        return cls(metric=metric.__name__, score=metric.score)


class MetricScores(BaseModel):
    metric: str
    scores: List[float]
    passes: int
    fails: int
    errors: int


class TraceMetricScores(BaseModel):
    agent: Dict[str, Dict[str, MetricScores]] = Field(default_factory=dict)
    tool: Dict[str, Dict[str, MetricScores]] = Field(default_factory=dict)
    retriever: Dict[str, Dict[str, MetricScores]] = Field(default_factory=dict)
    llm: Dict[str, Dict[str, MetricScores]] = Field(default_factory=dict)
    base: Dict[str, Dict[str, MetricScores]] = Field(default_factory=dict)


class PromptData(BaseModel):
    alias: Optional[str] = None
    version: Optional[str] = None
    text_template: Optional[str] = None
    messages_template: Optional[List[PromptMessage]] = None
    model_settings: Optional[ModelSettings] = None
    output_type: Optional[OutputType] = None
    interpolation_type: Optional[PromptInterpolationType] = None


class MetricsAverageDict:
    def __init__(self):
        self.metric_dict = {}
        self.metric_count = {}

    def add_metric(self, metric_name, score):
        if metric_name not in self.metric_dict:
            self.metric_dict[metric_name] = score
            self.metric_count[metric_name] = 1
        else:
            self.metric_dict[metric_name] += score
            self.metric_count[metric_name] += 1

    def get_average_metric_score(self):
        return [
            MetricScoreType(
                metric=metric,
                score=self.metric_dict[metric] / self.metric_count[metric],
            )
            for metric in self.metric_dict
        ]


class RemainingTestRun(BaseModel):
    testRunId: str
    test_cases: List[LLMApiTestCase] = Field(
        alias="testCases", default_factory=lambda: []
    )
    conversational_test_cases: List[ConversationalApiTestCase] = Field(
        alias="conversationalTestCases", default_factory=lambda: []
    )


class TestRun(BaseModel):
    test_file: Optional[str] = Field(
        None,
        alias="testFile",
    )
    test_cases: List[LLMApiTestCase] = Field(
        alias="testCases", default_factory=lambda: []
    )
    conversational_test_cases: List[ConversationalApiTestCase] = Field(
        alias="conversationalTestCases", default_factory=lambda: []
    )
    metrics_scores: List[MetricScores] = Field(
        default_factory=lambda: [], alias="metricsScores"
    )
    trace_metrics_scores: Optional[TraceMetricScores] = Field(
        None, alias="traceMetricsScores"
    )
    identifier: Optional[str] = None
    hyperparameters: Optional[Dict[str, Any]] = Field(None)
    prompts: Optional[List[PromptData]] = Field(None)
    test_passed: Optional[int] = Field(None, alias="testPassed")
    test_failed: Optional[int] = Field(None, alias="testFailed")
    run_duration: float = Field(0.0, alias="runDuration")
    evaluation_cost: Union[float, None] = Field(None, alias="evaluationCost")
    dataset_alias: Optional[str] = Field(None, alias="datasetAlias")
    dataset_id: Optional[str] = Field(None, alias="datasetId")

    def add_test_case(
        self, api_test_case: Union[LLMApiTestCase, ConversationalApiTestCase]
    ):
        if isinstance(api_test_case, ConversationalApiTestCase):
            self.conversational_test_cases.append(api_test_case)
        else:
            self.test_cases.append(api_test_case)

        if api_test_case.evaluation_cost is not None:
            if self.evaluation_cost is None:
                self.evaluation_cost = api_test_case.evaluation_cost
            else:
                self.evaluation_cost += api_test_case.evaluation_cost

    def set_dataset_properties(
        self,
        test_case: Union[LLMTestCase, ConversationalTestCase],
    ):
        if self.dataset_alias is None:
            self.dataset_alias = test_case._dataset_alias

        if self.dataset_id is None:
            self.dataset_id = test_case._dataset_id

    def sort_test_cases(self):
        self.test_cases.sort(
            key=lambda x: (x.order if x.order is not None else float("inf"))
        )
        # Optionally update order only if not already set
        highest_order = 0
        for test_case in self.test_cases:
            if test_case.order is None:
                test_case.order = highest_order
            highest_order = test_case.order + 1

        self.conversational_test_cases.sort(
            key=lambda x: (x.order if x.order is not None else float("inf"))
        )
        # Optionally update order only if not already set
        highest_order = 0
        for test_case in self.conversational_test_cases:
            if test_case.order is None:
                test_case.order = highest_order
            highest_order = test_case.order + 1

    def construct_metrics_scores(self) -> int:
        # Use a dict to aggregate scores, passes, and fails for each metric.
        metrics_dict: Dict[str, Dict[str, Any]] = {}
        # Add dict for trace metrics
        trace_metrics_dict: Dict[
            span_api_type_literals, Dict[str, Dict[str, Dict[str, Any]]]
        ] = {
            SpanApiType.AGENT.value: {},
            SpanApiType.TOOL.value: {},
            SpanApiType.RETRIEVER.value: {},
            SpanApiType.LLM.value: {},
            SpanApiType.BASE.value: {},
        }
        valid_scores = 0

        def process_metric_data(metric_data: MetricData):
            """
            Process and aggregate metric data for overall test metrics.

            Args:
                metric_data: The metric data to process
            """
            nonlocal valid_scores
            metric_name = metric_data.name
            score = metric_data.score
            success = metric_data.success

            if metric_name not in metrics_dict:
                metrics_dict[metric_name] = {
                    "scores": [],
                    "passes": 0,
                    "fails": 0,
                    "errors": 0,
                }

            metric_dict = metrics_dict[metric_name]

            if score is None or success is None:
                metric_dict["errors"] += 1
            else:
                valid_scores += 1
                metric_dict["scores"].append(score)
                if success:
                    metric_dict["passes"] += 1
                else:
                    metric_dict["fails"] += 1

        def process_span_metric_data(
            metric_data: MetricData,
            span_type: span_api_type_literals,
            span_name: str,
        ):
            """
            Process and aggregate metric data for a specific span.

            Args:
                metric_data: The metric data to process
                span_type: The type of span (agent, tool, retriever, llm, base)
                span_name: The name of the span
            """
            metric_name = metric_data.name
            score = metric_data.score
            success = metric_data.success

            if span_name not in trace_metrics_dict[span_type]:
                trace_metrics_dict[span_type][span_name] = {}

            if metric_name not in trace_metrics_dict[span_type][span_name]:
                trace_metrics_dict[span_type][span_name][metric_name] = {
                    "scores": [],
                    "passes": 0,
                    "fails": 0,
                    "errors": 0,
                }

            metric_dict = trace_metrics_dict[span_type][span_name][metric_name]

            if score is None or success is None:
                metric_dict["errors"] += 1
            else:
                metric_dict["scores"].append(score)
                if success:
                    metric_dict["passes"] += 1
                else:
                    metric_dict["fails"] += 1

        def process_spans(spans, span_type: span_api_type_literals):
            """
            Process all metrics for a list of spans of a specific type.

            Args:
                spans: List of spans to process
                span_type: The type of spans being processed
            """
            for span in spans:
                if span.metrics_data is not None:
                    for metric_data in span.metrics_data:
                        process_metric_data(metric_data)
                        process_span_metric_data(
                            metric_data, span_type, span.name
                        )

        # Process non-conversational test cases.
        for test_case in self.test_cases:
            if test_case.metrics_data is None:
                continue
            for metric_data in test_case.metrics_data:
                process_metric_data(metric_data)

            if test_case.trace is None:
                continue

            # Process all span types using the helper function
            process_spans(test_case.trace.agent_spans, SpanApiType.AGENT.value)
            process_spans(test_case.trace.tool_spans, SpanApiType.TOOL.value)
            process_spans(
                test_case.trace.retriever_spans, SpanApiType.RETRIEVER.value
            )
            process_spans(test_case.trace.llm_spans, SpanApiType.LLM.value)
            process_spans(test_case.trace.base_spans, SpanApiType.BASE.value)

        # Process conversational test cases.
        for convo_test_case in self.conversational_test_cases:
            if convo_test_case.metrics_data is not None:
                for metric_data in convo_test_case.metrics_data:
                    process_metric_data(metric_data)

        # Create MetricScores objects with the aggregated data.
        self.metrics_scores = [
            MetricScores(
                metric=metric,
                scores=data["scores"],
                passes=data["passes"],
                fails=data["fails"],
                errors=data["errors"],
            )
            for metric, data in metrics_dict.items()
        ]

        # Create a single TraceMetricScores object instead of a list
        trace_metrics_score = TraceMetricScores()
        has_span_metrics = False

        for span_type, spans in trace_metrics_dict.items():
            if not spans:  # Skip empty span types
                continue

            span_dict = {}
            for span_name, metrics in spans.items():
                span_dict[span_name] = {
                    metric_name: MetricScores(
                        metric=metric_name,
                        scores=metric_data["scores"],
                        passes=metric_data["passes"],
                        fails=metric_data["fails"],
                        errors=metric_data["errors"],
                    )
                    for metric_name, metric_data in metrics.items()
                }

            if span_dict:  # Only set if there are spans
                has_span_metrics = True
                setattr(trace_metrics_score, span_type, span_dict)

        # Set to None if no span metrics were found
        self.trace_metrics_scores = (
            trace_metrics_score if has_span_metrics else None
        )
        return valid_scores

    def calculate_test_passes_and_fails(self):
        test_passed = 0
        test_failed = 0
        for test_case in self.test_cases:
            if test_case.success is not None:
                if test_case.success:
                    test_passed += 1
                else:
                    test_failed += 1

        for test_case in self.conversational_test_cases:
            # we don't count for conversational messages success
            if test_case.success is not None:
                if test_case.success:
                    test_passed += 1
                else:
                    test_failed += 1

        self.test_passed = test_passed
        self.test_failed = test_failed

    def save(self, f):
        try:
            body = self.model_dump(by_alias=True, exclude_none=True)
        except AttributeError:
            body = self.dict(by_alias=True, exclude_none=True)
        json.dump(body, f, cls=TestRunEncoder)
        f.flush()
        os.fsync(f.fileno())
        return self

    @classmethod
    def load(cls, f):
        data: dict = json.load(f)
        return cls(**data)

    def guard_mllm_test_cases(self):
        for test_case in self.test_cases:
            if test_case.is_multimodal():
                raise ValueError(
                    "Unable to send multimodal test cases to Confident AI."
                )


class TestRunEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Enum):
            return obj.value
        return make_json_serializable(obj)


class TestRunManager:
    def __init__(self):
        self.test_run = None
        self.temp_file_path = TEMP_FILE_PATH
        self.save_to_disk = False
        self.disable_request = False

    def reset(self):
        self.test_run = None
        self.temp_file_path = TEMP_FILE_PATH
        self.save_to_disk = False
        self.disable_request = False

    def set_test_run(self, test_run: TestRun):
        self.test_run = test_run

    def create_test_run(
        self,
        identifier: Optional[str] = None,
        file_name: Optional[str] = None,
        disable_request: Optional[bool] = False,
    ):
        self.disable_request = disable_request
        test_run = TestRun(
            identifier=identifier,
            testFile=file_name,
            testCases=[],
            metricsScores=[],
            hyperparameters=None,
            testPassed=None,
            testFailed=None,
        )
        self.set_test_run(test_run)

        if self.save_to_disk:
            self.save_test_run(self.temp_file_path)

    def get_test_run(self, identifier: Optional[str] = None):
        if self.test_run is None:
            self.create_test_run(identifier=identifier)

        if portalocker and self.save_to_disk:
            try:
                with portalocker.Lock(
                    self.temp_file_path,
                    mode="r",
                    flags=portalocker.LOCK_SH | portalocker.LOCK_NB,
                ) as file:
                    loaded = self.test_run.load(file)
                    # only overwrite if loading actually worked
                    self.test_run = loaded
            except (
                FileNotFoundError,
                json.JSONDecodeError,
                portalocker.exceptions.LockException,
            ) as e:
                print(
                    f"Warning: Could not load test run from disk: {e}",
                    file=sys.stderr,
                )

        return self.test_run

    def save_test_run(self, path: str, save_under_key: Optional[str] = None):
        if portalocker and self.save_to_disk:
            try:
                # ensure parent directory exists
                parent = os.path.dirname(path)
                if parent:
                    os.makedirs(parent, exist_ok=True)

                with portalocker.Lock(path, mode="w") as file:
                    if save_under_key:
                        try:
                            test_run_data = self.test_run.model_dump(
                                by_alias=True, exclude_none=True
                            )
                        except AttributeError:
                            # Pydantic version below 2.0
                            test_run_data = self.test_run.dict(
                                by_alias=True, exclude_none=True
                            )
                        wrapper_data = {save_under_key: test_run_data}
                        json.dump(wrapper_data, file, cls=TestRunEncoder)
                        file.flush()
                        os.fsync(file.fileno())
                    else:
                        self.test_run.save(file)
            except portalocker.exceptions.LockException:
                pass

    def save_final_test_run_link(self, link: str):
        if portalocker:
            try:
                with portalocker.Lock(
                    LATEST_TEST_RUN_FILE_PATH, mode="w"
                ) as file:
                    json.dump({LATEST_TEST_RUN_LINK_KEY: link}, file)
                    file.flush()
                    os.fsync(file.fileno())
            except portalocker.exceptions.LockException:
                pass

    def update_test_run(
        self,
        api_test_case: Union[LLMApiTestCase, ConversationalApiTestCase],
        test_case: Union[LLMTestCase, ConversationalTestCase],
    ):
        if (
            api_test_case.metrics_data is not None
            and len(api_test_case.metrics_data) == 0
            and api_test_case.trace is None
        ):
            return

        if portalocker and self.save_to_disk:
            try:
                with portalocker.Lock(
                    self.temp_file_path,
                    mode="r+",
                    flags=portalocker.LOCK_EX,
                ) as file:
                    file.seek(0)
                    self.test_run = self.test_run.load(file)

                    # Update the test run object
                    self.test_run.add_test_case(api_test_case)
                    self.test_run.set_dataset_properties(test_case)

                    # Save the updated test run back to the file
                    file.seek(0)
                    file.truncate()
                    self.test_run.save(file)
            except (
                FileNotFoundError,
                json.JSONDecodeError,
                portalocker.exceptions.LockException,
            ) as e:
                print(
                    f"Warning: Could not update test run on disk: {e}",
                    file=sys.stderr,
                )
                if self.test_run is None:
                    # guarantee a valid in-memory run so the update can proceed.
                    # never destroy in-memory state on I/O failure.
                    self.create_test_run()
                self.test_run.add_test_case(api_test_case)
                self.test_run.set_dataset_properties(test_case)
        else:
            if self.test_run is None:
                self.create_test_run()

            self.test_run.add_test_case(api_test_case)
            self.test_run.set_dataset_properties(test_case)

    def clear_test_run(self):
        self.test_run = None

    @staticmethod
    def _calculate_success_rate(pass_count: int, fail_count: int) -> str:
        """Calculate success rate percentage or return error message."""
        total = pass_count + fail_count
        if total > 0:
            return str(round((100 * pass_count) / total, 2))
        return "Cannot display metrics for component-level evals, please run 'deepeval view' to see results on Confident AI."

    @staticmethod
    def _get_metric_status(metric_data: MetricData) -> str:
        """Get formatted status string for a metric."""
        if metric_data.error:
            return "[red]ERRORED[/red]"
        elif metric_data.success:
            return "[green]PASSED[/green]"
        return "[red]FAILED[/red]"

    @staticmethod
    def _format_metric_score(metric_data: MetricData) -> str:
        """Format metric score with evaluation details."""
        evaluation_model = metric_data.evaluation_model or "n/a"
        metric_score = (
            round(metric_data.score, 2)
            if metric_data.score is not None
            else None
        )

        return (
            f"{metric_score} "
            f"(threshold={metric_data.threshold}, "
            f"evaluation model={evaluation_model}, "
            f"reason={metric_data.reason}, "
            f"error={metric_data.error})"
        )

    @staticmethod
    def _should_skip_test_case(
        test_case, display: TestRunResultDisplay
    ) -> bool:
        """Determine if test case should be skipped based on display filter."""
        if display == TestRunResultDisplay.PASSING and not test_case.success:
            return True
        elif display == TestRunResultDisplay.FAILING and test_case.success:
            return True
        return False

    @staticmethod
    def _count_metric_results(
        metrics_data: List[MetricData],
    ) -> tuple[int, int]:
        """Count passing and failing metrics."""
        pass_count = 0
        fail_count = 0
        for metric_data in metrics_data:
            if metric_data.success:
                pass_count += 1
            else:
                fail_count += 1
        return pass_count, fail_count

    def _add_test_case_header_row(
        self,
        table: Table,
        test_case_name: str,
        pass_count: int,
        fail_count: int,
    ):
        """Add test case header row with name and success rate."""
        success_rate = self._calculate_success_rate(pass_count, fail_count)
        table.add_row(
            test_case_name,
            *[""] * 3,
            f"{success_rate}%",
        )

    def _add_metric_rows(self, table: Table, metrics_data: List[MetricData]):
        """Add metric detail rows to the table."""
        for metric_data in metrics_data:
            status = self._get_metric_status(metric_data)
            formatted_score = self._format_metric_score(metric_data)

            table.add_row(
                "",
                str(metric_data.name),
                formatted_score,
                status,
                "",
            )

    def _add_separator_row(self, table: Table):
        """Add empty separator row between test cases."""
        table.add_row(*[""] * len(table.columns))

    def display_results_table(
        self, test_run: TestRun, display: TestRunResultDisplay
    ):
        """Display test results in a formatted table."""

        table = Table(title="Test Results")
        column_config = dict(justify="left")
        column_names = [
            "Test case",
            "Metric",
            "Score",
            "Status",
            "Overall Success Rate",
        ]

        for name in column_names:
            table.add_column(name, **column_config)

        # Process regular test cases
        for index, test_case in enumerate(test_run.test_cases):
            if test_case.metrics_data is None or self._should_skip_test_case(
                test_case, display
            ):
                continue
            pass_count, fail_count = self._count_metric_results(
                test_case.metrics_data
            )
            self._add_test_case_header_row(
                table, test_case.name, pass_count, fail_count
            )
            self._add_metric_rows(table, test_case.metrics_data)

            if index < len(test_run.test_cases) - 1:
                self._add_separator_row(table)

        # Process conversational test cases
        for index, conversational_test_case in enumerate(
            test_run.conversational_test_cases
        ):
            if self._should_skip_test_case(conversational_test_case, display):
                continue

            conversational_test_case_name = conversational_test_case.name

            if conversational_test_case.turns:
                turns_table = Table(
                    title=f"Conversation - {conversational_test_case_name}",
                    show_header=True,
                    header_style="bold",
                )
                turns_table.add_column("#", justify="right", width=3)
                turns_table.add_column("Role", justify="left", width=10)

                # subtract fixed widths + borders and padding.
                # ~20 as a safe buffer
                details_max_width = max(
                    48, min(120, console.width - 3 - 10 - 20)
                )
                turns_table.add_column(
                    "Details",
                    justify="left",
                    overflow="fold",
                    max_width=details_max_width,
                )

                # truncate when too long
                tools_max_width = min(60, max(24, console.width // 3))
                turns_table.add_column(
                    "Tools",
                    justify="left",
                    no_wrap=True,
                    overflow="ellipsis",
                    max_width=tools_max_width,
                )

                sorted_turns = sorted(
                    conversational_test_case.turns, key=lambda t: t.order
                )

                for t in sorted_turns:
                    tools = t.tools_called or []
                    tool_names = ", ".join(tc.name for tc in tools)

                    # omit order, role and tools since we show them in a separate columns.
                    details = format_turn(
                        t,
                        include_tools_in_header=False,
                        include_order_role_in_header=False,
                    )

                    turns_table.add_row(
                        str(t.order),
                        t.role,
                        details,
                        shorten(tool_names, len_short()),
                    )

                console.print(turns_table)
            else:
                console.print(
                    f"[dim]No turns recorded for {conversational_test_case_name}.[/dim]"
                )
            if conversational_test_case.metrics_data is not None:
                pass_count, fail_count = self._count_metric_results(
                    conversational_test_case.metrics_data
                )
                self._add_test_case_header_row(
                    table, conversational_test_case.name, pass_count, fail_count
                )
                self._add_metric_rows(
                    table, conversational_test_case.metrics_data
                )

            if index < len(test_run.conversational_test_cases) - 1:
                self._add_separator_row(table)

            if index < len(test_run.test_cases) - 1:
                self._add_separator_row(table)

        table.add_row(
            "[bold red]Note: Use Confident AI with DeepEval to analyze failed test cases for more details[/bold red]",
            *[""] * (len(table.columns) - 1),
        )
        print(table)

    def post_test_run(self, test_run: TestRun) -> Optional[Tuple[str, str]]:
        if (
            len(test_run.test_cases) == 0
            and len(test_run.conversational_test_cases) == 0
        ):
            print("No test cases found, unable to upload to Confident AI.")
            return

        api = Api()

        is_conversational_run = len(test_run.conversational_test_cases) > 0
        all_test_cases_to_process = (
            test_run.conversational_test_cases
            if is_conversational_run
            else test_run.test_cases
        )

        custom_batch_size = os.getenv(CONFIDENT_TEST_CASE_BATCH_SIZE)
        if custom_batch_size and custom_batch_size.isdigit():
            BATCH_SIZE = int(custom_batch_size)
        else:
            BATCH_SIZE = 20 if is_conversational_run else 40

        initial_batch = all_test_cases_to_process[:BATCH_SIZE]
        remaining_test_cases_to_process = all_test_cases_to_process[BATCH_SIZE:]

        if len(remaining_test_cases_to_process) > 0:
            console.print(
                "Sending a large test run to Confident, this might take a bit longer than usual..."
            )

        ####################
        ### POST REQUEST ###
        ####################
        if is_conversational_run:
            test_run.conversational_test_cases = initial_batch
        else:
            test_run.test_cases = initial_batch

        try:
            test_run.prompts = None
            body = test_run.model_dump(by_alias=True, exclude_none=True)
        except AttributeError:
            # Pydantic version below 2.0
            body = test_run.dict(by_alias=True, exclude_none=True)

        json_str = json.dumps(body, cls=TestRunEncoder)
        body = json.loads(json_str)

        data, link = api.send_request(
            method=HttpMethods.POST,
            endpoint=Endpoints.TEST_RUN_ENDPOINT,
            body=body,
        )

        if not isinstance(data, dict) or "id" not in data:
            # try to show helpful details
            detail = None
            if isinstance(data, dict):
                detail = (
                    data.get("detail")
                    or data.get("message")
                    or data.get("error")
                )
            # fall back to repr for visibility
            raise RuntimeError(
                f"Confident API response missing 'id'. "
                f"detail={detail!r} raw={type(data).__name__}:{repr(data)[:500]}"
            )

        res = TestRunHttpResponse(
            id=data["id"],
        )

        ################################################
        ### Send the remaining test cases in batches ###
        ################################################
        total_remaining = len(remaining_test_cases_to_process)
        num_remaining_batches = (
            (total_remaining + BATCH_SIZE - 1) // BATCH_SIZE
            if total_remaining > 0
            else 0
        )

        for i in range(num_remaining_batches):
            start_index = i * BATCH_SIZE
            batch = remaining_test_cases_to_process[
                start_index : start_index + BATCH_SIZE
            ]

            if len(batch) == 0:
                break  # Should not happen with correct num_remaining_batches, but as a safeguard

            # Create RemainingTestRun with the correct list populated
            if is_conversational_run:
                remaining_test_run = RemainingTestRun(
                    testRunId=res.id,
                    testCases=[],  # This will be empty
                    conversationalTestCases=batch,
                )
            else:
                remaining_test_run = RemainingTestRun(
                    testRunId=res.id,
                    testCases=batch,
                    conversationalTestCases=[],  # This will be empty
                )

            body = None
            try:
                body = remaining_test_run.model_dump(
                    by_alias=True, exclude_none=True
                )
            except AttributeError:
                # Pydantic version below 2.0
                body = remaining_test_run.dict(by_alias=True, exclude_none=True)

            try:
                _, _ = api.send_request(
                    method=HttpMethods.PUT,
                    endpoint=Endpoints.TEST_RUN_ENDPOINT,
                    body=body,
                )
            except Exception as e:
                message = f"Unexpected error when sending some test cases. Incomplete test run available at {link}"
                raise Exception(message) from e

        console.print(
            "[rgb(5,245,141)]✓[/rgb(5,245,141)] Done 🎉! View results on "
            f"[link={link}]{link}[/link]"
        )
        self.save_final_test_run_link(link)
        open_browser(link)
        return link, res.id

    def save_test_run_locally(self):
        local_folder = os.getenv("DEEPEVAL_RESULTS_FOLDER")
        if local_folder:
            new_test_filename = datetime.datetime.now().strftime(
                "%Y%m%d_%H%M%S"
            )
            os.rename(self.temp_file_path, new_test_filename)
            if not os.path.exists(local_folder):
                os.mkdir(local_folder)
                shutil.copy(new_test_filename, local_folder)
                print(f"Results saved in {local_folder} as {new_test_filename}")
            elif os.path.isfile(local_folder):
                print(
                    f"""❌ Error: DEEPEVAL_RESULTS_FOLDER={local_folder} already exists and is a file.\nDetailed results won't be saved. Please specify a folder or an available path."""
                )
            else:
                shutil.copy(new_test_filename, local_folder)
                print(f"Results saved in {local_folder} as {new_test_filename}")
            os.remove(new_test_filename)

    def wrap_up_test_run(
        self,
        runDuration: float,
        display_table: bool = True,
        display: Optional[TestRunResultDisplay] = TestRunResultDisplay.ALL,
    ) -> Optional[Tuple[str, str]]:
        test_run = self.get_test_run()
        if test_run is None:
            print("Test Run is empty, please try again.")
            delete_file_if_exists(self.temp_file_path)
            return
        elif (
            len(test_run.test_cases) == 0
            and len(test_run.conversational_test_cases) == 0
        ):
            print("No test cases found, please try again.")
            delete_file_if_exists(self.temp_file_path)
            return

        valid_scores = test_run.construct_metrics_scores()
        if valid_scores == 0:
            print("All metrics errored for all test cases, please try again.")
            delete_file_if_exists(self.temp_file_path)
            delete_file_if_exists(
                global_test_run_cache_manager.temp_cache_file_name
            )
            return
        test_run.run_duration = runDuration
        test_run.calculate_test_passes_and_fails()
        test_run.sort_test_cases()

        if global_test_run_cache_manager.disable_write_cache is None:
            global_test_run_cache_manager.disable_write_cache = not bool(
                get_is_running_deepeval()
            )
        global_test_run_cache_manager.wrap_up_cached_test_run()

        if display_table:
            self.display_results_table(test_run, display)

        if test_run.hyperparameters is None:
            console.print(
                "\n[bold yellow]⚠ WARNING:[/bold yellow] No hyperparameters logged.\n"
                "» [bold blue][link=https://deepeval.com/docs/evaluation-prompts]Log hyperparameters[/link][/bold blue] to attribute prompts and models to your test runs.\n\n"
                + "=" * 80
            )
        else:
            if not test_run.prompts:
                console.print(
                    "\n[bold yellow]⚠ WARNING:[/bold yellow] No prompts logged.\n"
                    "» [bold blue][link=https://deepeval.com/docs/evaluation-prompts]Log prompts[/link][/bold blue] to evaluate and optimize your prompt templates and models.\n\n"
                    + "=" * 80
                )
            else:
                console.print("\n[bold green]✓ Prompts Logged[/bold green]\n")
                self._render_prompts_panels(prompts=test_run.prompts)

        self.save_test_run_locally()
        delete_file_if_exists(self.temp_file_path)
        if is_confident() and self.disable_request is False:
            return self.post_test_run(test_run)
        else:
            self.save_test_run(
                LATEST_TEST_RUN_FILE_PATH,
                save_under_key=LATEST_TEST_RUN_DATA_KEY,
            )
            console.print(
                f"\n\n[rgb(5,245,141)]✓[/rgb(5,245,141)] Evaluation completed 🎉! (time taken: {round(runDuration, 2)}s | token cost: {test_run.evaluation_cost} USD)\n"
                f"» Test Results ({test_run.test_passed + test_run.test_failed} total tests):\n",
                f"  » Pass Rate: {round((test_run.test_passed / (test_run.test_passed + test_run.test_failed)) * 100, 2)}% | Passed: [bold green]{test_run.test_passed}[/bold green] | Failed: [bold red]{test_run.test_failed}[/bold red]\n\n",
                "=" * 80,
                "\n\n» Want to share evals with your team, or a place for your test cases to live? ❤️ 🏡\n"
                "  » Run [bold]'deepeval view'[/bold] to analyze and save testing results on [rgb(106,0,255)]Confident AI[/rgb(106,0,255)].\n\n",
            )

    def get_latest_test_run_data(self) -> Optional[TestRun]:
        try:
            if os.path.exists(LATEST_TEST_RUN_FILE_PATH):
                with open(LATEST_TEST_RUN_FILE_PATH, "r") as file:
                    data = json.load(file)
                    return TestRun.model_validate(
                        data[LATEST_TEST_RUN_DATA_KEY]
                    )
        except (FileNotFoundError, json.JSONDecodeError, Exception):
            pass
        return None

    def get_latest_test_run_link(self) -> Optional[str]:
        try:
            if os.path.exists(LATEST_TEST_RUN_FILE_PATH):
                with open(LATEST_TEST_RUN_FILE_PATH, "r") as file:
                    data = json.load(file)
                    return data[LATEST_TEST_RUN_LINK_KEY]
        except (FileNotFoundError, json.JSONDecodeError, Exception):
            pass
        return None

    def _render_prompts_panels(self, prompts: List[PromptData]) -> None:

        def format_string(
            v, default="[dim]None[/dim]", color: Optional[str] = None
        ):
            formatted_string = str(v) if v not in (None, "", []) else default
            return (
                f"{formatted_string}"
                if color is None or v in (None, "", [])
                else f"[{color}]{formatted_string}[/]"
            )

        panels = []
        for prompt in prompts:
            lines = []
            p_type = (
                "messages"
                if prompt.messages_template
                else ("text" if prompt.text_template else "—")
            )
            if p_type:
                lines.append(f"type: {format_string(p_type, color='blue')}")
            if prompt.output_type:
                lines.append(
                    f"output_type: {format_string(prompt.output_type, color='blue')}"
                )
            if prompt.interpolation_type:
                lines.append(
                    f"interpolation_type: {format_string(prompt.interpolation_type, color='blue')}"
                )
            if prompt.model_settings:
                ms = prompt.model_settings
                settings_lines = [
                    "Model Settings:",
                    f"  – provider: {format_string(ms.provider, color='green')}",
                    f"  – name: {format_string(ms.name, color='green')}",
                    f"  – temperature: {format_string(ms.temperature, color='green')}",
                    f"  – max_tokens: {format_string(ms.max_tokens, color='green')}",
                    f"  – top_p: {format_string(ms.top_p, color='green')}",
                    f"  – frequency_penalty: {format_string(ms.frequency_penalty, color='green')}",
                    f"  – presence_penalty: {format_string(ms.presence_penalty, color='green')}",
                    f"  – stop_sequence: {format_string(ms.stop_sequence, color='green')}",
                    f"  – reasoning_effort: {format_string(ms.reasoning_effort, color='green')}",
                    f"  – verbosity: {format_string(ms.verbosity, color='green')}",
                ]
                lines.append("")
                lines.extend(settings_lines)
            title = f"{format_string(prompt.alias)}"
            if prompt.version:
                title += f" (v{prompt.version})"
            body = "\n".join(lines)
            panel = Panel(
                body,
                title=title,
                title_align="left",
                expand=False,
                padding=(1, 6, 1, 2),
            )
            panels.append(panel)

        if panels:
            console.print(Columns(panels, equal=False, expand=False))


global_test_run_manager = TestRunManager()



"""LLM evaluated metric based on the GEval framework: https://arxiv.org/pdf/2303.16634.pdf"""

from typing import Dict, Optional, List, Tuple, Union
from rich.progress import Progress

from deepeval.metrics import BaseArenaMetric
from deepeval.metrics.arena_g_eval.utils import format_arena_test_case
from deepeval.test_case import (
    LLMTestCaseParams,
    ArenaTestCase,
)
from deepeval.metrics.arena_g_eval.template import ArenaGEvalTemplate
from deepeval.utils import get_or_create_event_loop, prettify_list
from deepeval.metrics.utils import (
    check_arena_test_case_params,
    construct_verbose_logs,
    initialize_model,
    a_generate_with_schema_and_extract,
    generate_with_schema_and_extract,
)
from deepeval.models import DeepEvalBaseLLM
from deepeval.metrics.indicator import metric_progress_indicator
from deepeval.metrics.arena_g_eval.schema import (
    RewrittenReason,
    Winner,
    Steps,
)
from deepeval.metrics.g_eval.utils import (
    construct_g_eval_params_string,
    validate_criteria_and_evaluation_steps,
    number_evaluation_steps,
)
from deepeval.utils import update_pbar


class ArenaGEval(BaseArenaMetric):
    def __init__(
        self,
        name: str,
        evaluation_params: List[LLMTestCaseParams],
        criteria: Optional[str] = None,
        evaluation_steps: Optional[List[str]] = None,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        async_mode: bool = True,
        verbose_mode: bool = False,
        _include_g_eval_suffix: bool = True,
    ):
        validate_criteria_and_evaluation_steps(criteria, evaluation_steps)
        self.name = name
        self.evaluation_params = evaluation_params
        self.criteria = criteria
        self.model, self.using_native_model = initialize_model(model)
        self.evaluation_model = self.model.get_model_name()
        self.evaluation_steps = (
            evaluation_steps
            if evaluation_steps and len(evaluation_steps) > 0
            else None
        )
        self.async_mode = async_mode
        self.verbose_mode = verbose_mode
        self._include_g_eval_suffix = _include_g_eval_suffix

    def measure(
        self,
        test_case: ArenaTestCase,
        _show_indicator: bool = True,
        _progress: Optional[Progress] = None,
        _pbar_id: Optional[int] = None,
    ) -> str:
        check_arena_test_case_params(
            test_case,
            self.evaluation_params,
            self,
            self.model,
            test_case.multimodal,
        )
        self.evaluation_cost = 0 if self.using_native_model else None

        with metric_progress_indicator(self, _show_indicator=_show_indicator):
            if self.async_mode:
                loop = get_or_create_event_loop()
                loop.run_until_complete(
                    self.a_measure(
                        test_case,
                        _show_indicator=False,
                    )
                )
            else:
                self.evaluation_steps: List[str] = (
                    self._generate_evaluation_steps(test_case.multimodal)
                )
                if _progress:
                    update_pbar(_progress, _pbar_id)
                masked_winner, masked_reason, dummy_to_real_names = (
                    self._compare(test_case, test_case.multimodal)
                )
                if _progress:
                    update_pbar(_progress, _pbar_id)
                self.winner = dummy_to_real_names[masked_winner]
                self.reason = self._generate_rewritten_reason(
                    masked_reason, dummy_to_real_names
                )
                if _progress:
                    update_pbar(_progress, _pbar_id)
                self.success = True
                self.verbose_logs = construct_verbose_logs(
                    self,
                    steps=[
                        f"Criteria:\n{self.criteria}",
                        f"Evaluation Steps:\n{prettify_list(self.evaluation_steps)}",
                        f"Winner: {self.winner}",
                        f"Reason: {self.reason}",
                    ],
                )

            return self.winner

    async def a_measure(
        self,
        test_case: ArenaTestCase,
        _show_indicator: bool = True,
        _progress: Optional[Progress] = None,
        _pbar_id: Optional[int] = None,
    ) -> str:
        check_arena_test_case_params(
            test_case,
            self.evaluation_params,
            self,
            self.model,
            test_case.multimodal,
        )
        self.evaluation_cost = 0 if self.using_native_model else None

        with metric_progress_indicator(
            self,
            async_mode=True,
            _show_indicator=_show_indicator,
        ):
            self.evaluation_steps: List[str] = (
                await self._a_generate_evaluation_steps(test_case.multimodal)
            )
            if _progress:
                update_pbar(_progress, _pbar_id)
            masked_winner, masked_reason, dummy_to_real_names = (
                await self._a_compare(test_case, test_case.multimodal)
            )
            if _progress:
                update_pbar(_progress, _pbar_id)
            self.winner = dummy_to_real_names[masked_winner]
            self.reason = await self._a_generate_rewritten_reason(
                masked_reason, dummy_to_real_names
            )
            if _progress:
                update_pbar(_progress, _pbar_id)
            self.success = True
            self.verbose_logs = construct_verbose_logs(
                self,
                steps=[
                    f"Criteria:\n{self.criteria}",
                    f"Evaluation Steps:\n{prettify_list(self.evaluation_steps)}",
                    f"Winner: {self.winner}",
                    f"Reason: {self.reason}",
                ],
            )
            return self.winner

    async def _a_generate_evaluation_steps(self, multimodal: bool) -> List[str]:
        if self.evaluation_steps:
            return self.evaluation_steps

        g_eval_params_str = construct_g_eval_params_string(
            self.evaluation_params
        )
        prompt = ArenaGEvalTemplate.generate_evaluation_steps(
            criteria=self.criteria,
            parameters=g_eval_params_str,
            multimodal=multimodal,
        )

        return await a_generate_with_schema_and_extract(
            self,
            prompt,
            Steps,
            extract_schema=lambda s: s.steps,
            extract_json=lambda data: data["steps"],
        )

    def _generate_evaluation_steps(self, multimodal: bool) -> List[str]:
        if self.evaluation_steps:
            return self.evaluation_steps

        g_eval_params_str = construct_g_eval_params_string(
            self.evaluation_params
        )
        prompt = ArenaGEvalTemplate.generate_evaluation_steps(
            criteria=self.criteria,
            parameters=g_eval_params_str,
            multimodal=multimodal,
        )
        return generate_with_schema_and_extract(
            self,
            prompt,
            Steps,
            extract_schema=lambda s: s.steps,
            extract_json=lambda data: data["steps"],
        )

    async def _a_compare(
        self, test_case: ArenaTestCase, multimodal: bool
    ) -> Tuple[str, str, Dict[str, str]]:
        formatted_test_case, dummy_to_real_names = format_arena_test_case(
            self.evaluation_params, test_case
        )
        g_eval_params_str = construct_g_eval_params_string(
            self.evaluation_params
        )
        prompt = ArenaGEvalTemplate.generate_arena_winner(
            evaluation_steps=number_evaluation_steps(self.evaluation_steps),
            test_case_contents=formatted_test_case,
            parameters=g_eval_params_str,
            multimodal=multimodal,
        )

        return await a_generate_with_schema_and_extract(
            self,
            prompt,
            Winner,
            extract_schema=lambda s: (
                s.winner,
                s.reason,
                dummy_to_real_names,
            ),
            extract_json=lambda data: (
                data["winner"],
                data["reason"],
                dummy_to_real_names,
            ),
        )

    def _compare(
        self, test_case: ArenaTestCase, multimodal: bool
    ) -> Tuple[str, str, Dict[str, str]]:
        formatted_test_case, dummy_to_real_names = format_arena_test_case(
            self.evaluation_params, test_case
        )
        g_eval_params_str = construct_g_eval_params_string(
            self.evaluation_params
        )
        prompt = ArenaGEvalTemplate.generate_arena_winner(
            evaluation_steps=number_evaluation_steps(self.evaluation_steps),
            test_case_contents=formatted_test_case,
            parameters=g_eval_params_str,
            multimodal=multimodal,
        )
        return generate_with_schema_and_extract(
            self,
            prompt,
            Winner,
            extract_schema=lambda s: (
                s.winner,
                s.reason,
                dummy_to_real_names,
            ),
            extract_json=lambda data: (
                data["winner"],
                data["reason"],
                dummy_to_real_names,
            ),
        )

    async def _a_generate_rewritten_reason(
        self,
        reason: str,
        dummy_to_real_names: Dict[str, str],
    ) -> str:
        prompt = ArenaGEvalTemplate.rewrite_reason(
            reason=reason,
            dummy_to_real_names=dummy_to_real_names,
        )

        return await a_generate_with_schema_and_extract(
            self,
            prompt,
            RewrittenReason,
            extract_schema=lambda s: s.rewritten_reason,
            extract_json=lambda data: data["rewritten_reason"],
        )

    def _generate_rewritten_reason(
        self,
        reason: str,
        dummy_to_real_names: Dict[str, str],
    ) -> str:
        prompt = ArenaGEvalTemplate.rewrite_reason(
            reason=reason,
            dummy_to_real_names=dummy_to_real_names,
        )
        return generate_with_schema_and_extract(
            self,
            prompt,
            RewrittenReason,
            extract_schema=lambda s: s.rewritten_reason,
            extract_json=lambda data: data["rewritten_reason"],
        )

    def is_successful(self) -> bool:
        if self.error is not None:
            self.success = False
        return self.success

    @property
    def __name__(self):
        if self._include_g_eval_suffix:
            return f"{self.name} [Arena GEval]"
        else:
            return self.name



from enum import Enum
from dataclasses import dataclass, field
from pydantic import BaseModel, Field, ConfigDict
from typing import Any, Dict, List, Optional, Union, Literal
from rich.progress import Progress

from deepeval.utils import make_model_config

from deepeval.prompt.prompt import Prompt
from deepeval.test_case.llm_test_case import ToolCall
from deepeval.test_case import LLMTestCase
from deepeval.metrics import BaseMetric


class Message(BaseModel):
    role: str
    """To be displayed on the top of the message block."""

    type: Literal["tool_calls", "tool_output", "thinking", "default"] = (
        "default"
    )
    """Decides how the content is rendered."""

    content: Any
    """The content of the message."""


class TraceWorkerStatus(Enum):
    SUCCESS = "success"
    FAILURE = "failure"
    WARNING = "warning"


class SpanType(Enum):
    AGENT = "agent"
    LLM = "llm"
    RETRIEVER = "retriever"
    TOOL = "tool"


class TraceSpanStatus(Enum):
    SUCCESS = "SUCCESS"
    ERRORED = "ERRORED"
    IN_PROGRESS = "IN_PROGRESS"


class LlmToolCall(BaseModel):
    name: str
    args: Dict[str, Any]
    id: Optional[str] = None


class LlmOutput(BaseModel):
    role: str
    content: str
    tool_calls: Optional[List[LlmToolCall]] = None


class BaseSpan(BaseModel):
    model_config = make_model_config(arbitrary_types_allowed=True)

    uuid: str
    status: TraceSpanStatus
    children: List["BaseSpan"] = Field(default_factory=list)
    trace_uuid: str = Field(serialization_alias="traceUuid")
    parent_uuid: Optional[str] = Field(None, serialization_alias="parentUuid")
    start_time: float = Field(serialization_alias="startTime")
    end_time: Union[float, None] = Field(None, serialization_alias="endTime")
    name: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    input: Optional[Any] = None
    output: Optional[Any] = None
    error: Optional[str] = None
    llm_test_case: Optional[LLMTestCase] = None
    metrics: Optional[List[BaseMetric]] = None
    metric_collection: Optional[str] = None

    # Don't serialize these
    progress: Optional[Progress] = Field(None, exclude=True)
    pbar_callback_id: Optional[int] = Field(None, exclude=True)

    # additional test case parameters
    retrieval_context: Optional[List[str]] = Field(
        None, serialization_alias="retrievalContext"
    )
    context: Optional[List[str]] = Field(None, serialization_alias="context")
    expected_output: Optional[str] = Field(
        None, serialization_alias="expectedOutput"
    )
    tools_called: Optional[List[ToolCall]] = Field(
        None, serialization_alias="toolsCalled"
    )
    expected_tools: Optional[List[ToolCall]] = Field(
        None, serialization_alias="expectedTools"
    )


class AgentSpan(BaseSpan):
    name: str
    available_tools: List[str] = []
    agent_handoffs: List[str] = []


class LlmSpan(BaseSpan):

    model: Optional[str] = None
    prompt: Optional[Prompt] = None
    input_token_count: Optional[float] = Field(
        None, serialization_alias="inputTokenCount"
    )
    output_token_count: Optional[float] = Field(
        None, serialization_alias="outputTokenCount"
    )
    cost_per_input_token: Optional[float] = Field(
        None, serialization_alias="costPerInputToken"
    )
    cost_per_output_token: Optional[float] = Field(
        None, serialization_alias="costPerOutputToken"
    )
    token_intervals: Optional[Dict[float, str]] = Field(
        None, serialization_alias="tokenTimes"
    )

    # input_tools: Optional[List[ToolSchema]] = Field(None, serialization_alias="inputTools")
    # invocation_params: Optional[Dict[str, Any]] = Field(None, serialization_alias="invocationParams")
    # output_metadata: Optional[Dict[str, Any]] = Field(None, serialization_alias="outputMetadata")

    # for serializing `prompt`
    model_config = make_model_config(arbitrary_types_allowed=True)


class RetrieverSpan(BaseSpan):
    embedder: Optional[str] = None
    top_k: Optional[int] = Field(None, serialization_alias="topK")
    chunk_size: Optional[int] = Field(None, serialization_alias="chunkSize")


class ToolSpan(BaseSpan):
    name: str  # Required name for ToolSpan
    description: Optional[str] = None


class Trace(BaseModel):
    model_config = make_model_config(arbitrary_types_allowed=True)

    uuid: str = Field(serialization_alias="uuid")
    status: TraceSpanStatus
    root_spans: List[BaseSpan] = Field(serialization_alias="rootSpans")
    start_time: float = Field(serialization_alias="startTime")
    end_time: Union[float, None] = Field(None, serialization_alias="endTime")
    name: Optional[str] = None
    tags: Optional[List[str]] = None
    metadata: Optional[Dict[str, Any]] = None
    thread_id: Optional[str] = None
    user_id: Optional[str] = None
    input: Optional[Any] = None
    output: Optional[Any] = None
    metrics: Optional[List[BaseMetric]] = None
    metric_collection: Optional[str] = None

    # Don't serialize these
    confident_api_key: Optional[str] = Field(None, exclude=True)
    environment: str = Field(None, exclude=True)

    # additional test case parameters
    retrieval_context: Optional[List[str]] = Field(
        None, serialization_alias="retrievalContext"
    )
    context: Optional[List[str]] = Field(None, serialization_alias="context")
    expected_output: Optional[str] = Field(
        None, serialization_alias="expectedOutput"
    )
    tools_called: Optional[List[ToolCall]] = Field(
        None, serialization_alias="toolsCalled"
    )
    expected_tools: Optional[List[ToolCall]] = Field(
        None, serialization_alias="expectedTools"
    )


class TraceAttributes(BaseModel):
    name: Optional[str] = None
    tags: Optional[List[str]] = None
    metadata: Optional[Dict[str, Any]] = None
    thread_id: Optional[str] = None
    user_id: Optional[str] = None


@dataclass
class TestCaseMetricPair:
    test_case: LLMTestCase
    metrics: List[BaseMetric]
    hyperparameters: Optional[Dict[str, Any]] = field(default=None)



import textwrap
import json
from deepeval.tracing.utils import make_json_serializable


class PlanAdherenceTemplate:
    multimodal_rules = """
        --- MULTIMODAL INPUT RULES ---
        - Treat image content as factual evidence.
        - Only reference visual details that are explicitly and clearly visible.
        - Do not infer or guess objects, text, or details not visibly present.
        - If an image is unclear or ambiguous, mark uncertainty explicitly.
    """

    @staticmethod
    def extract_plan_from_trace(trace: dict) -> str:
        return textwrap.dedent(
            f"""You are a **systems analyst** evaluating an AI agent's execution trace.

                Your sole task is to extract the **explicit or clearly implied plan** the agent followed or intended to follow — *only if that plan is directly evidenced in the trace*.

                STRICT RULES TO FOLLOW:

                1. Source Evidence Requirement
                - Every plan step you include **must** be directly supported by explicit text in the trace.  
                - Acceptable evidence sources:
                    - `"reasoning"` or `"thought"` fields inside tool calls or function invocations.
                    - Explicit plan-like statements or lists written by the agent (e.g., “My plan is to…”).
                - If no evidence exists for a step, DO NOT infer or invent it.

                2. No Hallucination Policy 
                - You must *not* create or rephrase steps that aren't explicitly or strongly implied by the trace.  
                - If there is no coherent plan present, output an empty list.

                3. Focus on Intent, Not Outcomes
                - If the agent's plan is stated but execution differs, still extract the intended steps — but only if those intended steps are traceable.

                4. Granularity
                - Each step should represent a single distinct action or intention.
                - Avoid merging multiple intentions into one step or splitting one intention into multiple steps.

                5. Neutral Language
                - Reproduce the plan steps in **neutral, minimal paraphrasing**.  
                - Do not interpret motivation, quality, or success of actions.

                {PlanAdherenceTemplate.multimodal_rules}

                OUTPUT FORMAT:

                Return a JSON object with exactly this structure:
                {{
                    "plan": [
                        "step 1",
                        "step 2",
                        ...
                    ]
                }}

                If no plan is evidenced in the trace, return:
                {{
                    "plan": []
                }}

                Do not include commentary, confidence scores, or explanations.

                TRACE:

                {json.dumps(trace, indent=2, default=str)}

                JSON:
            """
        )

    @staticmethod
    def evaluate_adherence(
        user_task: str, agent_plan: str, execution_trace: dict
    ) -> str:
        return textwrap.dedent(
            f"""You are an **adversarial plan adherence evaluator**. Your goal is to assign the **lowest justifiable score** based on how strictly the agent's actions in the execution trace align with its declared plan.

                INPUTS:

                - **User Task:** The original request or objective.
                - **Agent Plan:** The explicit step-by-step plan the agent was supposed to follow.
                - **Execution Trace:** A detailed record of all agent actions, reasoning, tool calls, and outputs.

                EVALUATION OBJECTIVE:

                Determine whether the agent **exactly and exclusively** followed its plan.  
                You are not evaluating success, correctness, or usefulness — **only plan obedience**.

                Assume **non-adherence by default** unless clear, direct evidence in the trace proves that 
                each planned step was executed *as written* and *no additional actions occurred*.

                ### STRICT ADHERENCE RULES

                1. Step Verification
                - Every step in the plan must correspond to a **verifiable, explicit** action or reasoning entry in the trace.
                - Each step must appear in the same logical order as the plan.
                - If a step is missing, only implied, or ambiguous, treat as **not followed**.

                2. No Extraneous Actions
                - If the trace includes **any** major action, tool call, or reasoning segment not clearly present in the plan, immediately lower the score to as low as possible.
                - Extra or unnecessary steps are considered serious violations.

                3. Order Consistency
                - If the agent performed steps in a different order than the plan specifies, the score must be close to 0, regardless of other alignment.

                4. Completeness
                - If even one planned step is missing, skipped, or only partially reflected in the trace, the score must be lowest possible.

                5. Ambiguity Handling
                - If it is unclear whether a trace action corresponds to a plan step, treat that step as **not executed**.
                - When uncertain, assign the **lower score**.

                6. Focus Exclusively on Plan Compliance
                - Ignore task success, reasoning quality, or correctness of outcomes. 
                - Evaluate *only* whether the trace reflects the exact plan execution.

                {PlanAdherenceTemplate.multimodal_rules}


                SCORING SCALE

                - **1.0 — Perfect adherence**
                - Every planned step is explicitly and verifiably present in the trace, in correct order.
                - No skipped or added steps.
                - No ambiguity in matching.

                - **0.75 — Strong adherence**
                - All or nearly all steps are executed in order.
                - At most one minor deviation (e.g., a trivial reordering or minor redundant step) that does not change the plan’s structure.

                - **0.5 — Partial adherence**
                - Some steps clearly match, but others are missing, out of order, or replaced.
                - At least one extra or ambiguous action appears.
                - *This should be the highest score possible when there are any deviations.*

                - **0.25 — Weak adherence**
                - Only a few steps from the plan appear in the trace, or multiple extraneous actions occur.
                - The structure or sequence of the plan is mostly lost.

                - **0.0 — No adherence**
                - The trace shows little or no resemblance to the plan.
                - Steps are ignored, replaced, or executed in an entirely different order.

                Always err toward the **lower score** when evidence is partial, ambiguous, or contradictory.

                OUTPUT FORMAT:

                Return a JSON object with exactly this structure:

                {{
                    "score": 0.0,
                    "reason": "1-3 concise, factual sentences citing specific matched, missing, or extra steps."
                }}

                Requirements for `"reason"`:
                - Reference specific plan step numbers or phrases.
                - Mention concrete trace evidence of mismatches or additions.
                - Avoid subjective adjectives (e.g., “mostly”, “close”, “reasonable”).
                - Be precise and neutral.

                ---

                INPUTS:

                User Task:
                {user_task}

                Agent Plan:
                {agent_plan}

                Execution Trace:
                {json.dumps(execution_trace, indent=2, default=make_json_serializable)}

                ---

                JSON:
            """
        )



from __future__ import annotations
import json
import random
from typing import List, Optional, Tuple, Union

from deepeval.errors import DeepEvalError
from deepeval.optimizer.utils import (
    validate_int_in_range,
    validate_instance,
)
from deepeval.optimizer.configs import (
    MutationConfig,
    MutationTargetType,
)
from deepeval.prompt.api import PromptType, PromptMessage
from deepeval.prompt.prompt import Prompt


##################
# Common Helpers #
##################
def _summarize_prompt_for_rewrite(old_prompt: Prompt, max_chars: int) -> str:
    """
    Produce a human-readable summary of the current prompt for the
    rewriter instruction block.

    - For TEXT prompts, this is just `text_template`.
    - For LIST prompts, this is a numbered list of (role, content) lines.
    """

    # LIST prompts: show each message with its role.
    if old_prompt.type is PromptType.LIST and old_prompt.messages_template:
        lines: List[str] = []
        for message_index, message in enumerate(old_prompt.messages_template):
            role = message.role or ""
            content = message.content or ""
            lines.append(f"[{message_index+1}] ({role}) {content}")
        combined = "\n".join(lines)
        return combined[:max_chars]

    # Since it is not a LIST prompt, just use text_template.
    text = old_prompt.text_template or ""
    return text[:max_chars]


def _select_list_target_index(
    messages: List[PromptMessage],
    config: MutationConfig,
    random_state: random.Random,
) -> int:
    """
    Select which list message index to rewrite, based on PromptListMutationConfig.

    Rules:
    - Start with all indices in scope.
    - If target_role is set, restrict candidates to messages with that role
      (case insensitive). If no messages match, fall back to all indices.
    - target_type:
        * FIRST:       pick the first candidate index.
        * RANDOM:      pick a candidate via random_state.choice(candidates).
        * FIXED_INDEX: use target_index when valid (and consistent with role
                       filter), otherwise fall back to the first candidate.
    """
    if not messages:
        raise DeepEvalError(
            "Rewriter._select_list_target_index expected at least one "
            "message, but received an empty message list."
        )

    validate_instance(
        component="Rewriter._select_list_target_index",
        param_name="target_type",
        value=config.target_type,
        expected_types=MutationTargetType,
    )

    messages_length = len(messages)
    candidate_indices = list(range(messages_length))

    # Optional case insensitive role restriction
    if config.target_role:
        target_role_lower = config.target_role.lower()
        filtered = [
            index
            for index, message in enumerate(messages)
            if (message.role or "").lower() == target_role_lower
        ]
        if filtered:
            candidate_indices = filtered

    target_type = config.target_type

    if target_type is MutationTargetType.RANDOM:
        return random_state.choice(candidate_indices)

    if target_type is MutationTargetType.FIXED_INDEX:
        index = validate_int_in_range(
            component="Rewriter._select_list_target_index",
            param_name="target_index",
            value=int(config.target_index),
            min_inclusive=0,
            max_exclusive=len(candidate_indices),
        )
        return candidate_indices[index]

    # if you got this error it means that a new PromptListMutationTargetType was added,
    # but not handled above
    raise DeepEvalError(
        "Rewriter._select_list_target_index received unsupported "
        f"target_type={target_type!r}. Expected RANDOM or FIXED_INDEX."
    )


def _apply_rewritten_prompt(
    old_prompt: Prompt,
    new_text: str,
    random_state: random.Random,
    list_mutation_config: Optional[MutationConfig] = None,
) -> Prompt:
    """
    Apply the rewritten text to a Prompt, preserving representation:

    - For TEXT prompts, update `text_template`.
    - For LIST prompts, rewrite the content of a single message while
      keeping the number of messages the same.
    - Preserve additonal Prompt meta such as `label` and `interpolation_type`
    """
    if not new_text:
        return old_prompt

    if old_prompt.type is PromptType.LIST and old_prompt.messages_template:
        messages = old_prompt.messages_template
        config = list_mutation_config or MutationConfig()

        target_index = _select_list_target_index(
            messages=messages,
            config=config,
            random_state=random_state,
        )

        new_messages: List[PromptMessage] = []
        for message_index, message in enumerate(messages):
            if message_index == target_index:
                # Preserve the original role; do not inject a new one.
                new_messages.append(
                    PromptMessage(
                        role=message.role,
                        content=new_text,
                    )
                )
            else:
                new_messages.append(message)

        new_prompt = Prompt(
            alias=old_prompt.alias,
            text_template=None,
            messages_template=new_messages,
            model_settings=old_prompt.model_settings,
            output_type=old_prompt.output_type,
            output_schema=old_prompt.output_schema,
        )

    else:
        # Since it is not LIST, it must be TEXT type
        new_prompt = Prompt(
            alias=old_prompt.alias,
            text_template=new_text,
            model_settings=old_prompt.model_settings,
            output_type=old_prompt.output_type,
            output_schema=old_prompt.output_schema,
        )

    new_prompt.label = old_prompt.label
    new_prompt.interpolation_type = old_prompt.interpolation_type
    return new_prompt


def _compose_prompt_messages(system_message: str, user_message: str) -> str:
    """
    Join system and user messages into a single prompt string.
    Strips surrounding whitespace from each part; if the system message is
    empty or absent, returns just the user message.
    """
    system_text = (system_message or "").strip()
    user_text = (user_message or "").strip()
    return f"{system_text}\n\n{user_text}" if system_text else user_text


def _normalize_llm_output_to_text(
    result: Union[str, Tuple[Union[str, dict], float], dict],
) -> str:
    """
    Convert a DeepEval LLM generate() / a_generate() result to a clean string.

    Accepted inputs:
      - str                        -> returned as trimmed
      - (str|dict, float_cost)     -> first element extracted and normalized
      - dict (e.g. JSON mode)      -> JSON serialized with ensure_ascii=False

    Fallback: if serialization fails, str(value).strip() is used.
    """
    output_value: Union[str, dict]
    if isinstance(result, tuple):
        output_value = result[0]
    else:
        output_value = result

    if isinstance(output_value, str):
        return output_value.strip()

    try:
        return json.dumps(output_value, ensure_ascii=False)
    except Exception:
        return str(output_value).strip()



from typing import Optional, Tuple, Union, Dict
from openai import OpenAI, AsyncOpenAI
from pydantic import BaseModel, SecretStr

from deepeval.errors import DeepEvalError
from deepeval.config.settings import get_settings
from deepeval.models.llms.utils import trim_and_load_json
from deepeval.models.utils import (
    require_costs,
    require_secret_api_key,
)
from deepeval.models import DeepEvalBaseLLM
from deepeval.models.retry_policy import (
    create_retry_decorator,
    sdk_retries_for,
)
from deepeval.constants import ProviderSlug as PS
from deepeval.models.llms.constants import DEEPSEEK_MODELS_DATA
from deepeval.utils import require_param


# consistent retry rules
retry_deepseek = create_retry_decorator(PS.DEEPSEEK)


class DeepSeekModel(DeepEvalBaseLLM):
    def __init__(
        self,
        model: Optional[str] = None,
        api_key: Optional[str] = None,
        temperature: Optional[float] = None,
        cost_per_input_token: Optional[float] = None,
        cost_per_output_token: Optional[float] = None,
        generation_kwargs: Optional[Dict] = None,
        **kwargs,
    ):
        settings = get_settings()

        model = model or settings.DEEPSEEK_MODEL_NAME

        if temperature is not None:
            temperature = float(temperature)
        elif settings.TEMPERATURE is not None:
            temperature = settings.TEMPERATURE
        else:
            temperature = 0.0

        cost_per_input_token = (
            cost_per_input_token
            if cost_per_input_token is not None
            else settings.DEEPSEEK_COST_PER_INPUT_TOKEN
        )
        cost_per_output_token = (
            cost_per_output_token
            if cost_per_output_token is not None
            else settings.DEEPSEEK_COST_PER_OUTPUT_TOKEN
        )

        if api_key is not None:
            # keep it secret, keep it safe from serializings, logging and alike
            self.api_key: Optional[SecretStr] = SecretStr(api_key)
        else:
            self.api_key = settings.DEEPSEEK_API_KEY

        self.base_url = "https://api.deepseek.com"

        # validation
        model = require_param(
            model,
            provider_label="DeepSeekModel",
            env_var_name="DEEPSEEK_MODEL_NAME",
            param_hint="model",
        )

        if temperature < 0:
            raise DeepEvalError("Temperature must be >= 0.")

        self.model_data = DEEPSEEK_MODELS_DATA.get(model)
        self.temperature = temperature

        cost_per_input_token, cost_per_output_token = require_costs(
            self.model_data,
            model,
            "DEEPSEEK_COST_PER_INPUT_TOKEN",
            "DEEPSEEK_COST_PER_OUTPUT_TOKEN",
            cost_per_input_token,
            cost_per_output_token,
        )
        self.model_data.input_price = cost_per_input_token
        self.model_data.output_price = cost_per_output_token

        # Keep sanitized kwargs for client call to strip legacy keys
        self.kwargs = kwargs
        self.kwargs.pop("temperature", None)

        self.generation_kwargs = dict(generation_kwargs or {})
        self.generation_kwargs.pop("temperature", None)

        super().__init__(model)

    ###############################################
    # Other generate functions
    ###############################################

    @retry_deepseek
    def generate(
        self, prompt: str, schema: Optional[BaseModel] = None
    ) -> Tuple[Union[str, BaseModel], float]:

        client = self.load_model(async_mode=False)
        if schema:
            completion = client.chat.completions.create(
                model=self.name,
                messages=[{"role": "user", "content": prompt}],
                response_format={"type": "json_object"},
                temperature=self.temperature,
                **self.generation_kwargs,
            )
            json_output = trim_and_load_json(
                completion.choices[0].message.content
            )
            cost = self.calculate_cost(
                completion.usage.prompt_tokens,
                completion.usage.completion_tokens,
            )
            return schema.model_validate(json_output), cost
        else:
            completion = client.chat.completions.create(
                model=self.name,
                messages=[{"role": "user", "content": prompt}],
                **self.generation_kwargs,
            )
            output = completion.choices[0].message.content
            cost = self.calculate_cost(
                completion.usage.prompt_tokens,
                completion.usage.completion_tokens,
            )
            return output, cost

    @retry_deepseek
    async def a_generate(
        self, prompt: str, schema: Optional[BaseModel] = None
    ) -> Tuple[Union[str, BaseModel], float]:

        client = self.load_model(async_mode=True)
        if schema:
            completion = await client.chat.completions.create(
                model=self.name,
                messages=[{"role": "user", "content": prompt}],
                response_format={"type": "json_object"},
                temperature=self.temperature,
                **self.generation_kwargs,
            )
            json_output = trim_and_load_json(
                completion.choices[0].message.content
            )
            cost = self.calculate_cost(
                completion.usage.prompt_tokens,
                completion.usage.completion_tokens,
            )
            return schema.model_validate(json_output), cost
        else:
            completion = await client.chat.completions.create(
                model=self.name,
                messages=[{"role": "user", "content": prompt}],
                **self.generation_kwargs,
            )
            output = completion.choices[0].message.content
            cost = self.calculate_cost(
                completion.usage.prompt_tokens,
                completion.usage.completion_tokens,
            )
            return output, cost

    ###############################################
    # Utilities
    ###############################################

    def calculate_cost(
        self,
        input_tokens: int,
        output_tokens: int,
    ) -> float:
        input_cost = input_tokens * self.model_data.input_price
        output_cost = output_tokens * self.model_data.output_price
        return input_cost + output_cost

    ###############################################
    # Capabilities
    ###############################################

    def supports_log_probs(self) -> Union[bool, None]:
        return self.model_data.supports_log_probs

    def supports_temperature(self) -> Union[bool, None]:
        return self.model_data.supports_temperature

    def supports_multimodal(self) -> Union[bool, None]:
        return self.model_data.supports_multimodal

    def supports_structured_outputs(self) -> Union[bool, None]:
        return self.model_data.supports_structured_outputs

    def supports_json_mode(self) -> Union[bool, None]:
        return self.model_data.supports_json

    ###############################################
    # Model
    ###############################################

    def load_model(self, async_mode: bool = False):
        if not async_mode:
            return self._build_client(OpenAI)
        return self._build_client(AsyncOpenAI)

    def _client_kwargs(self) -> Dict:
        kwargs = dict(self.kwargs or {})
        # if we are managing retries with Tenacity, force SDK retries off to avoid double retries.
        # if the user opts into SDK retries for "deepseek" via DEEPEVAL_SDK_RETRY_PROVIDERS, honor it.
        if not sdk_retries_for(PS.DEEPSEEK):
            kwargs["max_retries"] = 0
        return kwargs

    def _build_client(self, cls):
        api_key = require_secret_api_key(
            self.api_key,
            provider_label="DeepSeek",
            env_var_name="DEEPSEEK_API_KEY",
            param_hint="`api_key` to DeepSeekModel(...)",
        )

        kw = dict(
            api_key=api_key,
            base_url=self.base_url,
            **self._client_kwargs(),
        )
        try:
            return cls(**kw)
        except TypeError as e:
            # In case an older OpenAI client doesn’t accept max_retries, drop it and retry.
            if "max_retries" in str(e):
                kw.pop("max_retries", None)
                return cls(**kw)
            raise

    def get_model_name(self):
        return f"{self.name} (Deepseek)"



"""File for handling API key"""

import os
import json
import logging

from enum import Enum
from typing import Union

from .constants import KEY_FILE, HIDDEN_DIR


logger = logging.getLogger(__name__)


SECRET_KEYS = {
    # General providers
    "OPENAI_API_KEY",
    "ANTHROPIC_API_KEY",
    # Azure OpenAI
    "AZURE_OPENAI_API_KEY",
    # Google / Gemini
    "GOOGLE_API_KEY",
    # xAI Grok
    "GROK_API_KEY",
    # Moonshot
    "MOONSHOT_API_KEY",
    # DeepSeek
    "DEEPSEEK_API_KEY",
    # LiteLLM
    "LITELLM_API_KEY",
    # Local gateways (if any require keys)
    "LOCAL_MODEL_API_KEY",
    "LOCAL_EMBEDDING_API_KEY",
}

_WARNED_SECRET_KEYS = set()


class KeyValues(Enum):
    # Confident AI
    API_KEY = "api_key"
    CONFIDENT_REGION = "confident_region"
    # Cache
    LAST_TEST_RUN_LINK = "last_test_run_link"
    LAST_TEST_RUN_DATA = "last_test_run_data"


class ModelKeyValues(Enum):
    # General
    TEMPERATURE = "TEMPERATURE"
    # Azure Open AI
    AZURE_OPENAI_API_KEY = "AZURE_OPENAI_API_KEY"
    AZURE_OPENAI_ENDPOINT = "AZURE_OPENAI_ENDPOINT"
    OPENAI_API_VERSION = "OPENAI_API_VERSION"
    AZURE_DEPLOYMENT_NAME = "AZURE_DEPLOYMENT_NAME"
    AZURE_MODEL_NAME = "AZURE_MODEL_NAME"
    AZURE_MODEL_VERSION = "AZURE_MODEL_VERSION"
    USE_AZURE_OPENAI = "USE_AZURE_OPENAI"
    # Local Model
    LOCAL_MODEL_NAME = "LOCAL_MODEL_NAME"
    LOCAL_MODEL_BASE_URL = "LOCAL_MODEL_BASE_URL"
    LOCAL_MODEL_API_KEY = "LOCAL_MODEL_API_KEY"
    LOCAL_MODEL_FORMAT = "LOCAL_MODEL_FORMAT"
    USE_LOCAL_MODEL = "USE_LOCAL_MODEL"
    # Gemini
    USE_GEMINI_MODEL = "USE_GEMINI_MODEL"
    GEMINI_MODEL_NAME = "GEMINI_MODEL_NAME"
    GOOGLE_API_KEY = "GOOGLE_API_KEY"
    GOOGLE_GENAI_USE_VERTEXAI = "GOOGLE_GENAI_USE_VERTEXAI"
    GOOGLE_CLOUD_PROJECT = "GOOGLE_CLOUD_PROJECT"
    GOOGLE_CLOUD_LOCATION = "GOOGLE_CLOUD_LOCATION"
    GOOGLE_SERVICE_ACCOUNT_KEY = "GOOGLE_SERVICE_ACCOUNT_KEY"
    # LiteLLM
    USE_LITELLM = "USE_LITELLM"
    LITELLM_MODEL_NAME = "LITELLM_MODEL_NAME"
    LITELLM_API_KEY = "LITELLM_API_KEY"
    LITELLM_API_BASE = "LITELLM_API_BASE"
    # OpenAI
    USE_OPENAI_MODEL = "USE_OPENAI_MODEL"
    OPENAI_MODEL_NAME = "OPENAI_MODEL_NAME"
    OPENAI_COST_PER_INPUT_TOKEN = "OPENAI_COST_PER_INPUT_TOKEN"
    OPENAI_COST_PER_OUTPUT_TOKEN = "OPENAI_COST_PER_OUTPUT_TOKEN"
    OPENAI_API_KEY = "OPENAI_API_KEY"
    # Moonshot
    USE_MOONSHOT_MODEL = "USE_MOONSHOT_MODEL"
    MOONSHOT_MODEL_NAME = "MOONSHOT_MODEL_NAME"
    MOONSHOT_API_KEY = "MOONSHOT_API_KEY"
    # Grok
    USE_GROK_MODEL = "USE_GROK_MODEL"
    GROK_MODEL_NAME = "GROK_MODEL_NAME"
    GROK_API_KEY = "GROK_API_KEY"
    # DeepSeek
    USE_DEEPSEEK_MODEL = "USE_DEEPSEEK_MODEL"
    DEEPSEEK_MODEL_NAME = "DEEPSEEK_MODEL_NAME"
    DEEPSEEK_API_KEY = "DEEPSEEK_API_KEY"


class EmbeddingKeyValues(Enum):
    # Azure OpenAI
    USE_AZURE_OPENAI_EMBEDDING = "USE_AZURE_OPENAI_EMBEDDING"
    # Azure OpenAI
    AZURE_EMBEDDING_MODEL_NAME = "AZURE_EMBEDDING_MODEL_NAME"
    AZURE_EMBEDDING_DEPLOYMENT_NAME = "AZURE_EMBEDDING_DEPLOYMENT_NAME"

    # Local
    USE_LOCAL_EMBEDDINGS = "USE_LOCAL_EMBEDDINGS"
    LOCAL_EMBEDDING_MODEL_NAME = "LOCAL_EMBEDDING_MODEL_NAME"
    LOCAL_EMBEDDING_BASE_URL = "LOCAL_EMBEDDING_BASE_URL"
    LOCAL_EMBEDDING_API_KEY = "LOCAL_EMBEDDING_API_KEY"


class KeyFileHandler:
    def __init__(self):
        self.data = {}

    def _ensure_dir(self):
        os.makedirs(HIDDEN_DIR, exist_ok=True)

    def write_key(
        self, key: Union[KeyValues, ModelKeyValues, EmbeddingKeyValues], value
    ):
        """Appends or updates data in the hidden file"""

        # hard stop on secrets: never write to disk
        if key.value in SECRET_KEYS:
            logger.warning(
                f"{key} is blacklisted, refusing to persist. Keep your secrets in .env or .env.local instead"
            )
            return

        try:
            with open(f"{HIDDEN_DIR}/{KEY_FILE}", "r") as f:
                # Load existing data
                try:
                    self.data = json.load(f)
                except json.JSONDecodeError:
                    # Handle corrupted JSON file
                    self.data = {}
        except FileNotFoundError:
            # If file doesn't exist, start with an empty dictionary
            self.data = {}

        # Update the data with the new key-value pair
        self.data[key.value] = value

        # Write the updated data back to the file
        self._ensure_dir()
        with open(f"{HIDDEN_DIR}/{KEY_FILE}", "w") as f:
            json.dump(self.data, f)

    def fetch_data(
        self, key: Union[KeyValues, ModelKeyValues, EmbeddingKeyValues]
    ):
        """Fetches the data from the hidden file.
        NOTE: secrets in this file are deprecated; prefer env/.env."""
        try:
            with open(f"{HIDDEN_DIR}/{KEY_FILE}", "r") as f:
                try:
                    self.data = json.load(f)
                except json.JSONDecodeError:
                    # Handle corrupted JSON file
                    self.data = {}
        except FileNotFoundError:
            # Handle the case when the file doesn't exist
            self.data = {}

        value = self.data.get(key.value)

        # Deprecation: warn only if we're actually returning a secret
        if (
            value is not None
            and key.value in SECRET_KEYS
            and key.value not in _WARNED_SECRET_KEYS
        ):
            logger.warning(
                f"Reading secret '{key.value}' from legacy {HIDDEN_DIR}/{KEY_FILE}. "
                "Persisting API keys in plaintext is deprecated. "
                "Move this to your environment (.env / .env.local). "
                "This fallback will be removed in a future release."
            )
            _WARNED_SECRET_KEYS.add(key.value)

        return value

    def remove_key(
        self, key: Union[KeyValues, ModelKeyValues, EmbeddingKeyValues]
    ):
        """Removes the specified key from the data."""
        try:
            with open(f"{HIDDEN_DIR}/{KEY_FILE}", "r") as f:
                try:
                    self.data = json.load(f)
                except json.JSONDecodeError:
                    # Handle corrupted JSON file
                    self.data = {}
            self.data.pop(key.value, None)  # Remove the key if it exists
            self._ensure_dir()
            with open(f"{HIDDEN_DIR}/{KEY_FILE}", "w") as f:
                json.dump(self.data, f)
        except FileNotFoundError:
            # Handle the case when the file doesn't exist
            pass  # No action needed if the file doesn't exist


KEY_FILE_HANDLER = KeyFileHandler()



from typing import Callable, List
from functools import wraps


from deepeval.openai.extractors import (
    safe_extract_output_parameters,
    safe_extract_input_parameters,
    InputParameters,
    OutputParameters,
)
from deepeval.test_case.llm_test_case import ToolCall
from deepeval.tracing.context import (
    current_trace_context,
    update_current_span,
    update_llm_span,
)
from deepeval.tracing import observe
from deepeval.tracing.trace_context import current_llm_context

# Store original methods for safety and potential unpatching
_ORIGINAL_METHODS = {}
_OPENAI_PATCHED = False


def patch_openai_classes():
    """Monkey patch OpenAI resource classes directly."""
    global _OPENAI_PATCHED

    # Single guard - if already patched, return immediately
    if _OPENAI_PATCHED:
        return

    try:
        from openai.resources.chat.completions import (
            Completions,
            AsyncCompletions,
        )

        # Store original methods before patching
        if hasattr(Completions, "create"):
            _ORIGINAL_METHODS["Completions.create"] = Completions.create
            Completions.create = _create_sync_wrapper(
                Completions.create, is_completion_method=True
            )

        if hasattr(Completions, "parse"):
            _ORIGINAL_METHODS["Completions.parse"] = Completions.parse
            Completions.parse = _create_sync_wrapper(
                Completions.parse, is_completion_method=True
            )

        if hasattr(AsyncCompletions, "create"):
            _ORIGINAL_METHODS["AsyncCompletions.create"] = (
                AsyncCompletions.create
            )
            AsyncCompletions.create = _create_async_wrapper(
                AsyncCompletions.create, is_completion_method=True
            )

        if hasattr(AsyncCompletions, "parse"):
            _ORIGINAL_METHODS["AsyncCompletions.parse"] = AsyncCompletions.parse
            AsyncCompletions.parse = _create_async_wrapper(
                AsyncCompletions.parse, is_completion_method=True
            )

    except ImportError:
        pass

    try:
        from openai.resources.responses import Responses, AsyncResponses

        if hasattr(Responses, "create"):
            _ORIGINAL_METHODS["Responses.create"] = Responses.create
            Responses.create = _create_sync_wrapper(
                Responses.create, is_completion_method=False
            )

        if hasattr(AsyncResponses, "create"):
            _ORIGINAL_METHODS["AsyncResponses.create"] = AsyncResponses.create
            AsyncResponses.create = _create_async_wrapper(
                AsyncResponses.create, is_completion_method=False
            )

    except ImportError:
        pass

    # Set flag at the END after successful patching
    _OPENAI_PATCHED = True


def _create_sync_wrapper(original_method, is_completion_method: bool):
    """Create a wrapper for sync methods - called ONCE during patching."""

    @wraps(original_method)
    def method_wrapper(self, *args, **kwargs):
        bound_method = original_method.__get__(self, type(self))
        patched = _patch_sync_openai_client_method(
            orig_method=bound_method, is_completion_method=is_completion_method
        )
        return patched(*args, **kwargs)

    return method_wrapper


def _create_async_wrapper(original_method, is_completion_method: bool):
    """Create a wrapper for async methods - called ONCE during patching."""

    @wraps(original_method)
    async def method_wrapper(self, *args, **kwargs):
        bound_method = original_method.__get__(self, type(self))
        patched = _patch_async_openai_client_method(
            orig_method=bound_method, is_completion_method=is_completion_method
        )
        return await patched(*args, **kwargs)

    return method_wrapper


def _patch_async_openai_client_method(
    orig_method: Callable,
    is_completion_method: bool = False,
):
    @wraps(orig_method)
    async def patched_async_openai_method(*args, **kwargs):
        input_parameters: InputParameters = safe_extract_input_parameters(
            is_completion_method, kwargs
        )

        llm_context = current_llm_context.get()

        @observe(
            type="llm",
            model=input_parameters.model,
            metrics=llm_context.metrics,
            metric_collection=llm_context.metric_collection,
        )
        async def llm_generation(*args, **kwargs):
            response = await orig_method(*args, **kwargs)
            output_parameters = safe_extract_output_parameters(
                is_completion_method, response, input_parameters
            )
            _update_all_attributes(
                input_parameters,
                output_parameters,
                llm_context.expected_tools,
                llm_context.expected_output,
                llm_context.context,
                llm_context.retrieval_context,
            )

            return response

        return await llm_generation(*args, **kwargs)

    return patched_async_openai_method


def _patch_sync_openai_client_method(
    orig_method: Callable,
    is_completion_method: bool = False,
):
    @wraps(orig_method)
    def patched_sync_openai_method(*args, **kwargs):
        input_parameters: InputParameters = safe_extract_input_parameters(
            is_completion_method, kwargs
        )

        llm_context = current_llm_context.get()

        @observe(
            type="llm",
            model=input_parameters.model,
            metrics=llm_context.metrics,
            metric_collection=llm_context.metric_collection,
        )
        def llm_generation(*args, **kwargs):
            response = orig_method(*args, **kwargs)
            output_parameters = safe_extract_output_parameters(
                is_completion_method, response, input_parameters
            )
            _update_all_attributes(
                input_parameters,
                output_parameters,
                llm_context.expected_tools,
                llm_context.expected_output,
                llm_context.context,
                llm_context.retrieval_context,
            )

            return response

        return llm_generation(*args, **kwargs)

    return patched_sync_openai_method


def _update_all_attributes(
    input_parameters: InputParameters,
    output_parameters: OutputParameters,
    expected_tools: List[ToolCall],
    expected_output: str,
    context: List[str],
    retrieval_context: List[str],
):
    """Update span and trace attributes with input/output parameters."""
    update_current_span(
        input=input_parameters.messages,
        output=output_parameters.output or output_parameters.tools_called,
        tools_called=output_parameters.tools_called,
        # attributes to be added
        expected_output=expected_output,
        expected_tools=expected_tools,
        context=context,
        retrieval_context=retrieval_context,
    )

    llm_context = current_llm_context.get()

    update_llm_span(
        input_token_count=output_parameters.prompt_tokens,
        output_token_count=output_parameters.completion_tokens,
        prompt=llm_context.prompt,
    )

    __update_input_and_output_of_current_trace(
        input_parameters, output_parameters
    )


def __update_input_and_output_of_current_trace(
    input_parameters: InputParameters, output_parameters: OutputParameters
):

    current_trace = current_trace_context.get()
    if current_trace:
        if current_trace.input is None:
            current_trace.input = (
                input_parameters.input or input_parameters.messages
            )

        if current_trace.output is None:
            current_trace.output = output_parameters.output

    return


def unpatch_openai_classes():
    """Restore OpenAI resource classes to their original state."""
    global _OPENAI_PATCHED

    # If not patched, nothing to do
    if not _OPENAI_PATCHED:
        return

    try:
        from openai.resources.chat.completions import (
            Completions,
            AsyncCompletions,
        )

        # Restore original methods for Completions
        if "Completions.create" in _ORIGINAL_METHODS:
            Completions.create = _ORIGINAL_METHODS["Completions.create"]

        if "Completions.parse" in _ORIGINAL_METHODS:
            Completions.parse = _ORIGINAL_METHODS["Completions.parse"]

        # Restore original methods for AsyncCompletions
        if "AsyncCompletions.create" in _ORIGINAL_METHODS:
            AsyncCompletions.create = _ORIGINAL_METHODS[
                "AsyncCompletions.create"
            ]

        if "AsyncCompletions.parse" in _ORIGINAL_METHODS:
            AsyncCompletions.parse = _ORIGINAL_METHODS["AsyncCompletions.parse"]

    except ImportError:
        pass

    try:
        from openai.resources.responses import Responses, AsyncResponses

        # Restore original methods for Responses
        if "Responses.create" in _ORIGINAL_METHODS:
            Responses.create = _ORIGINAL_METHODS["Responses.create"]

        # Restore original methods for AsyncResponses
        if "AsyncResponses.create" in _ORIGINAL_METHODS:
            AsyncResponses.create = _ORIGINAL_METHODS["AsyncResponses.create"]

    except ImportError:
        pass

    # Reset the patched flag
    _OPENAI_PATCHED = False



from .test_run import (
    TestRun,
    global_test_run_manager,
    TEMP_FILE_PATH,
    LATEST_TEST_RUN_FILE_PATH,
    LATEST_TEST_RUN_DATA_KEY,
    LATEST_TEST_RUN_LINK_KEY,
    LLMApiTestCase,
    ConversationalApiTestCase,
    TestRunManager,
    PromptData,
)

from .hooks import on_test_run_end, invoke_test_run_end_hook
from .api import MetricData, TurnApi
from .hyperparameters import log_hyperparameters


__all__ = [
    "TestRun",
    "global_test_run_manager",
    "TEMP_FILE_PATH",
    "LATEST_TEST_RUN_FILE_PATH",
    "LATEST_TEST_RUN_DATA_KEY",
    "LATEST_TEST_RUN_LINK_KEY",
    "LLMApiTestCase",
    "ConversationalApiTestCase",
    "TestRunManager",
    "on_test_run_end",
    "invoke_test_run_end_hook",
    "MetricData",
    "TurnApi",
    "log_hyperparameters",
]



from contextlib import contextmanager
import logging
import os
import socket
import sys
import uuid
import sentry_sdk
from enum import Enum
from typing import List, Dict
import requests
from deepeval.config.settings import get_settings
from deepeval.constants import LOGIN_PROMPT, HIDDEN_DIR, KEY_FILE
from posthog import Posthog


class Feature(Enum):
    REDTEAMING = "redteaming"
    SYNTHESIZER = "synthesizer"
    EVALUATION = "evaluation"
    COMPONENT_EVALUATION = "component_evaluation"
    GUARDRAIL = "guardrail"
    BENCHMARK = "benchmark"
    CONVERSATION_SIMULATOR = "conversation_simulator"
    UNKNOWN = "unknown"
    TRACING_INTEGRATION = "tracing_integration"


TELEMETRY_DATA_FILE = ".deepeval_telemetry.txt"
TELEMETRY_PATH = os.path.join(HIDDEN_DIR, TELEMETRY_DATA_FILE)

#########################################################
### Telemetry HELPERS ###################################
#########################################################


def telemetry_opt_out():
    return get_settings().DEEPEVAL_TELEMETRY_OPT_OUT


def blocked_by_firewall():
    try:
        socket.create_connection(("www.google.com", 80))
        return False
    except OSError:
        return True


def get_anonymous_public_ip():
    try:
        response = requests.get("https://api.ipify.org", timeout=5)
        if response.status_code == 200:
            return response.text
    except requests.RequestException:
        pass
    return None


#########################################################
### Move Folders ########################################
#########################################################
if not telemetry_opt_out():
    if os.path.exists(KEY_FILE) and not os.path.isdir(HIDDEN_DIR):
        temp_deepeval_file_name = ".deepeval_temp"
        os.rename(KEY_FILE, temp_deepeval_file_name)
        os.makedirs(HIDDEN_DIR, exist_ok=True)
        os.rename(temp_deepeval_file_name, os.path.join(HIDDEN_DIR, KEY_FILE))

    os.makedirs(HIDDEN_DIR, exist_ok=True)

    if os.path.exists(TELEMETRY_DATA_FILE):
        os.rename(TELEMETRY_DATA_FILE, TELEMETRY_PATH)

    if os.path.exists(".deepeval-cache.json"):
        os.rename(".deepeval-cache.json", f"{HIDDEN_DIR}/.deepeval-cache.json")

    if os.path.exists(".temp_test_run_data.json"):
        os.rename(
            ".temp_test_run_data.json", f"{HIDDEN_DIR}/.temp_test_run_data.json"
        )

#########################################################
### Telemetry Config ####################################
#########################################################

anonymous_public_ip = None

if not telemetry_opt_out():
    from opentelemetry import trace
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import BatchSpanProcessor
    from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
        OTLPSpanExporter,
    )

    anonymous_public_ip = get_anonymous_public_ip()
    sentry_sdk.init(
        dsn="https://5ef587d58109ee45d6544f3657efdd1f@o4506098477236224.ingest.sentry.io/4506098479136768",
        profiles_sample_rate=1.0,
        traces_sample_rate=1.0,  # For performance monitoring
        send_default_pii=False,  # Don't send personally identifiable information
        attach_stacktrace=False,  # Don't attach stack traces to messages
        default_integrations=False,  # Disable Sentry's default integrations
    )

    # Set up the Tracer Provider
    trace.set_tracer_provider(TracerProvider())
    tracer_provider = trace.get_tracer_provider()

    # New Relic License Key and OTLP Endpoint
    NEW_RELIC_LICENSE_KEY = "1711c684db8a30361a7edb0d0398772cFFFFNRAL"
    NEW_RELIC_OTLP_ENDPOINT = "https://otlp.nr-data.net:4317"
    otlp_exporter = OTLPSpanExporter(
        endpoint=NEW_RELIC_OTLP_ENDPOINT,
        headers={"api-key": NEW_RELIC_LICENSE_KEY},
    )

    # Add the OTLP exporter to the span processor
    span_processor = BatchSpanProcessor(otlp_exporter)
    tracer_provider.add_span_processor(span_processor)

    logging.getLogger("opentelemetry.exporter.otlp").setLevel(logging.CRITICAL)

    # Create a tracer for your application
    tracer = trace.get_tracer(__name__)

    # Initialize PostHog
    posthog = Posthog(
        project_api_key="phc_IXvGRcscJJoIb049PtjIZ65JnXQguOUZ5B5MncunFdB",
        host="https://us.i.posthog.com",
    )


if (
    get_settings().ERROR_REPORTING
    and not blocked_by_firewall()
    and not telemetry_opt_out()
):

    def handle_exception(exc_type, exc_value, exc_traceback):
        print({"exc_type": exc_type, "exc_value": exc_value})
        sentry_sdk.capture_exception(exc_value)
        sys.__excepthook__(exc_type, exc_value, exc_traceback)

    sys.excepthook = handle_exception


def is_running_in_jupyter_notebook():
    try:
        from IPython import get_ipython

        if "IPKernelApp" in get_ipython().config:
            return True
    except Exception:
        pass
    return False


IS_RUNNING_IN_JUPYTER = (
    "jupyter" if is_running_in_jupyter_notebook() else "other"
)

#########################################################
### Context Managers ####################################
#########################################################


@contextmanager
def capture_evaluation_run(type: str):
    if telemetry_opt_out():
        yield
    else:
        # data
        event = f"Ran {type}"
        distinct_id = get_unique_id()
        feature = (
            Feature.COMPONENT_EVALUATION
            if event == "Ran traceable evaluate()"
            else Feature.EVALUATION
        )
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
        }
        if feature == Feature.EVALUATION:
            properties["feature_status.evaluation"] = get_feature_status(
                feature
            )
        elif feature == Feature.COMPONENT_EVALUATION:
            properties["feature_status.component_evaluation"] = (
                get_feature_status(feature)
            )
        set_last_feature(feature)
        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            yield span


@contextmanager
def capture_recommend_metrics():
    if telemetry_opt_out():
        yield
    else:
        # data
        event = "Recommend"
        distinct_id = get_unique_id()
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
        }
        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            yield span


@contextmanager
def capture_metric_type(
    metric_name: str, async_mode: bool, in_component: bool, _track: bool = True
):
    if telemetry_opt_out():
        yield
    else:
        # data
        event = metric_name
        distinct_id = get_unique_id()
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
            "async_mode": async_mode,
            "in_component": int(in_component),
        }
        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            yield span


@contextmanager
def capture_synthesizer_run(
    method: str, max_generations: int, num_evolutions: int, evolutions: Dict
):
    if telemetry_opt_out():
        yield
    else:
        # data
        event = "Invoked synthesizer"
        distinct_id = get_unique_id()
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
            "feature_status.synthesizer": get_feature_status(
                Feature.SYNTHESIZER
            ),
            "method": method,
            "max_generations": max_generations,
            "num_evolutions": num_evolutions,
            **{f"evolution.{evol.value}": 1 for evol in evolutions},
        }
        set_last_feature(Feature.SYNTHESIZER)
        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            yield span


@contextmanager
def capture_conversation_simulator_run(num_conversations: int):
    if telemetry_opt_out():
        yield
    else:
        # data
        event = "Invoked conversation simulator"
        distinct_id = get_unique_id()
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
            "feature_status.conversation_simulator": get_feature_status(
                Feature.CONVERSATION_SIMULATOR
            ),
            "num_conversations": num_conversations,
        }
        set_last_feature(Feature.CONVERSATION_SIMULATOR)
        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            yield span


@contextmanager
def capture_guardrails(guards: List[str]):
    if telemetry_opt_out():
        yield
    else:
        event = "Ran guardrails"
        distinct_id = get_unique_id()
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
            "feature_status.guardrail": get_feature_status(Feature.GUARDRAIL),
            **{f"vulnerability.{guard}": 1 for guard in guards},
        }
        set_last_feature(Feature.GUARDRAIL)
        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            yield span


@contextmanager
def capture_benchmark_run(benchmark: str, num_tasks: int):
    if telemetry_opt_out():
        yield
    else:
        event = "Ran benchmark"
        distinct_id = get_unique_id()
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
            "feature_status.benchmark": get_feature_status(Feature.BENCHMARK),
            "benchmark": benchmark,
            "num_tasks": num_tasks,
        }
        set_last_feature(Feature.BENCHMARK)
        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            yield span


@contextmanager
def capture_login_event():
    if telemetry_opt_out():
        yield
    else:
        event = "Login"
        distinct_id = get_unique_id()
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
            "last_feature": get_last_feature().value,
            "completed": True,
            "login_prompt": LOGIN_PROMPT,
        }
        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            yield span


@contextmanager
def capture_view_event():
    if telemetry_opt_out():
        yield
    else:
        event = "View"
        distinct_id = get_unique_id()
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
            "last_feature": get_last_feature().value,
            "completed": True,
            "login_prompt": LOGIN_PROMPT,
        }
        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            yield span


@contextmanager
def capture_pull_dataset():
    if telemetry_opt_out():
        yield
    else:
        event = "Pull"
        distinct_id = get_unique_id()
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
        }
        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            yield span


# track metrics that are components and metrics that aren't components


# number of traces
@contextmanager
def capture_send_trace():
    if telemetry_opt_out():
        yield
    else:
        event = "Send Trace"
        distinct_id = get_unique_id()
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
        }
        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            yield span


# tracing integration
@contextmanager
def capture_tracing_integration(integration_name: str):
    if telemetry_opt_out():
        yield
    else:
        event = f"Tracing Integration: deepeval.integrations.{integration_name}"
        distinct_id = get_unique_id()
        properties = {
            "logged_in_with": get_logged_in_with(),
            "environment": IS_RUNNING_IN_JUPYTER,
            "user.status": get_status(),
            "user.unique_id": get_unique_id(),
            "user.public_ip": (
                anonymous_public_ip if anonymous_public_ip else "Unknown"
            ),
            "feature_status.tracing_integration": get_feature_status(
                Feature.TRACING_INTEGRATION
            ),
        }
        set_last_feature(Feature.TRACING_INTEGRATION)

        # capture posthog
        posthog.capture(
            distinct_id=distinct_id, event=event, properties=properties
        )
        # capture new relic
        with tracer.start_as_current_span(event) as span:
            for property, value in properties.items():
                span.set_attribute(property, value)
            # OTEL/New Relic filtering attributes
            span.set_attribute("integration.name", integration_name)
            yield span


#########################################################
### Helper Functions s####################################
#########################################################


def read_telemetry_file() -> dict:
    """Reads the telemetry data file and returns the key-value pairs as a dictionary."""
    if not os.path.exists(TELEMETRY_PATH):
        return {}
    with open(TELEMETRY_PATH, "r") as file:
        lines = file.readlines()
    data = {}
    for line in lines:
        key, _, value = line.strip().partition("=")
        data[key] = value
    return data


def write_telemetry_file(data: dict):
    """Writes the given key-value pairs to the telemetry data file."""
    # respect opt out
    if telemetry_opt_out():
        return

    # ensure directory exists before write
    os.makedirs(HIDDEN_DIR, exist_ok=True)
    with open(TELEMETRY_PATH, "w") as file:
        for key, value in data.items():
            file.write(f"{key}={value}\n")


def get_status() -> str:
    """Gets the status from the telemetry file."""
    data = read_telemetry_file()
    return data.get("DEEPEVAL_STATUS", "new")


def get_unique_id() -> str:
    """Gets or generates a unique ID and updates the telemetry file."""
    # respect opt out
    if telemetry_opt_out():
        return "telemetry-opted-out"
    data = read_telemetry_file()
    unique_id = data.get("DEEPEVAL_ID")
    if not unique_id:
        unique_id = str(uuid.uuid4())
        data["DEEPEVAL_ID"] = unique_id
        data["DEEPEVAL_STATUS"] = "new"
    else:
        data["DEEPEVAL_STATUS"] = "old"
    write_telemetry_file(data)
    return unique_id


def get_last_feature() -> Feature:
    """Gets the last feature from the telemetry file."""
    data = read_telemetry_file()
    last_feature = data.get("DEEPEVAL_LAST_FEATURE")
    if last_feature and last_feature in Feature._value2member_map_:
        return Feature(last_feature)
    return Feature.UNKNOWN


def set_last_feature(feature: Feature):
    """Sets the last feature in the telemetry file."""
    if feature not in Feature:
        raise ValueError(f"Invalid feature: {feature}")
    data = read_telemetry_file()
    data["DEEPEVAL_LAST_FEATURE"] = feature.value
    feature_status_key = f"DEEPEVAL_{feature.value.upper()}_STATUS"
    data[feature_status_key] = "old"
    write_telemetry_file(data)


def get_feature_status(feature: Feature) -> str:
    """Gets the status of a feature ('new' or 'old') from the telemetry file."""
    data = read_telemetry_file()
    feature_status_key = f"DEEPEVAL_{feature.value.upper()}_STATUS"
    return data.get(feature_status_key, "new")


def set_logged_in_with(logged_in_with: str):
    data = read_telemetry_file()
    data["LOGGED_IN_WITH"] = logged_in_with
    write_telemetry_file(data)


def get_logged_in_with():
    data = read_telemetry_file()
    return data.get("LOGGED_IN_WITH", "NA")



import logging
from typing import Optional, Any, Union, Tuple
import aiohttp
import requests
from enum import Enum
import os
from tenacity import (
    retry,
    wait_exponential_jitter,
    retry_if_exception_type,
    RetryCallState,
)
from pydantic import SecretStr

import deepeval
from deepeval.key_handler import KEY_FILE_HANDLER, KeyValues
from deepeval.confident.types import ApiResponse, ConfidentApiError
from deepeval.config.settings import get_settings


CONFIDENT_API_KEY_ENV_VAR = "CONFIDENT_API_KEY"
DEEPEVAL_BASE_URL = "https://deepeval.confident-ai.com"
DEEPEVAL_BASE_URL_EU = "https://eu.deepeval.confident-ai.com"
API_BASE_URL = "https://api.confident-ai.com"
API_BASE_URL_EU = "https://eu.api.confident-ai.com"
retryable_exceptions = requests.exceptions.SSLError


def get_base_api_url():
    s = get_settings()
    if s.CONFIDENT_BASE_URL:
        base_url = s.CONFIDENT_BASE_URL.rstrip("/")
        return base_url
    region = KEY_FILE_HANDLER.fetch_data(KeyValues.CONFIDENT_REGION)
    if region == "EU":
        return API_BASE_URL_EU
    else:
        return API_BASE_URL


def get_confident_api_key() -> Optional[str]:
    s = get_settings()
    key: Optional[SecretStr] = s.CONFIDENT_API_KEY or s.API_KEY
    return key.get_secret_value() if key else None


def set_confident_api_key(api_key: Optional[str]) -> None:
    """
    - Always updates runtime (os.environ) via settings.edit()
    - If DEEPEVAL_DEFAULT_SAVE is set, also persists to dotenv
    - Never writes secrets to the legacy JSON keystore (your Settings logic already skips secrets)
    """
    s = get_settings()
    save = (
        s.DEEPEVAL_DEFAULT_SAVE or None
    )  # e.g. "dotenv" or "dotenv:/path/.env"

    # If you *only* want runtime changes unless a default save is present:
    if save is None:
        with s.edit(persist=False):
            s.CONFIDENT_API_KEY = SecretStr(api_key) if api_key else None
            s.API_KEY = SecretStr(api_key) if api_key else None
    else:
        # Respect default save: update runtime + write to dotenv, but not JSON
        with s.edit(save=save, persist=None):
            s.CONFIDENT_API_KEY = SecretStr(api_key) if api_key else None
            s.API_KEY = SecretStr(api_key) if api_key else None


def is_confident():
    confident_api_key = get_confident_api_key()
    return confident_api_key is not None


def log_retry_error(retry_state: RetryCallState):
    exception = retry_state.outcome.exception()
    logging.error(
        f"Confident AI Error: {exception}. Retrying: {retry_state.attempt_number} time(s)..."
    )


class HttpMethods(Enum):
    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    PUT = "PUT"


class Endpoints(Enum):
    DATASET_ALIAS_ENDPOINT = "/v1/datasets/:alias"
    DATASET_ALIAS_QUEUE_ENDPOINT = "/v1/datasets/:alias/queue"

    TEST_RUN_ENDPOINT = "/v1/test-run"
    EXPERIMENT_ENDPOINT = "/v1/experiment"
    METRIC_DATA_ENDPOINT = "/v1/metric-data"
    TRACES_ENDPOINT = "/v1/traces"
    ANNOTATIONS_ENDPOINT = "/v1/annotations"
    PROMPTS_VERSION_ID_ENDPOINT = "/v1/prompts/:alias/versions/:versionId"
    PROMPTS_LABEL_ENDPOINT = "/v1/prompts/:alias/labels/:label"
    PROMPTS_ENDPOINT = "/v1/prompts"
    PROMPTS_VERSIONS_ENDPOINT = "/v1/prompts/:alias/versions"
    SIMULATE_ENDPOINT = "/v1/simulate"
    EVALUATE_ENDPOINT = "/v1/evaluate"

    EVALUATE_THREAD_ENDPOINT = "/v1/evaluate/threads/:threadId"
    EVALUATE_TRACE_ENDPOINT = "/v1/evaluate/traces/:traceUuid"
    EVALUATE_SPAN_ENDPOINT = "/v1/evaluate/spans/:spanUuid"


class Api:
    def __init__(self, api_key: Optional[str] = None):
        if api_key is None:
            api_key = get_confident_api_key()

        if not api_key:
            raise ValueError(
                f"No Confident API key found. Please run `deepeval login` or set the {CONFIDENT_API_KEY_ENV_VAR} environment variable in the CLI."
            )

        self.api_key = api_key
        self._headers = {
            "Content-Type": "application/json",
            "CONFIDENT-API-KEY": api_key,
            "X-DeepEval-Version": deepeval.__version__,
        }
        self.base_api_url = get_base_api_url()

    @staticmethod
    @retry(
        wait=wait_exponential_jitter(initial=1, exp_base=2, jitter=2, max=10),
        retry=retry_if_exception_type(retryable_exceptions),
        after=log_retry_error,
    )
    def _http_request(
        method: str, url: str, headers=None, json=None, params=None
    ):
        session = requests.Session()
        return session.request(
            method=method,
            url=url,
            headers=headers,
            json=json,
            params=params,
            verify=True,  # SSL verification is always enabled
        )

    def _handle_response(
        self, response_data: Union[dict, Any]
    ) -> Tuple[Any, Optional[str]]:
        if not isinstance(response_data, dict):
            return response_data, None

        try:
            api_response = ApiResponse(**response_data)
        except Exception:
            return response_data, None

        if api_response.deprecated:
            deprecation_msg = "You are using a deprecated API endpoint. Please update your deepeval version."
            if api_response.link:
                deprecation_msg += f" See: {api_response.link}"
            logging.warning(deprecation_msg)

        if not api_response.success:
            error_message = api_response.error or "Request failed"
            raise ConfidentApiError(error_message, api_response.link)

        return api_response.data, api_response.link

    def send_request(
        self,
        method: HttpMethods,
        endpoint: Endpoints,
        body=None,
        params=None,
        url_params=None,
    ) -> Tuple[Any, Optional[str]]:
        url = f"{self.base_api_url}{endpoint.value}"

        # Replace URL parameters if provided
        if url_params:
            for key, value in url_params.items():
                placeholder = f":{key}"
                if placeholder in url:
                    url = url.replace(placeholder, str(value))

        res = self._http_request(
            method=method.value,
            url=url,
            headers=self._headers,
            json=body,
            params=params,
        )

        if res.status_code == 200:
            try:
                response_data = res.json()
                return self._handle_response(response_data)
            except ValueError:
                return res.text, None
        else:
            try:
                error_data = res.json()
                return self._handle_response(error_data)
            except (ValueError, ConfidentApiError) as e:
                if isinstance(e, ConfidentApiError):
                    raise e
                error_message = (
                    error_data.get("error", res.text)
                    if "error_data" in locals()
                    else res.text
                )
                raise Exception(error_message)

    async def a_send_request(
        self,
        method: HttpMethods,
        endpoint: Endpoints,
        body=None,
        params=None,
        url_params=None,
    ) -> Tuple[Any, Optional[str]]:
        url = f"{self.base_api_url}{endpoint.value}"

        if url_params:
            for key, value in url_params.items():
                placeholder = f":{key}"
                if placeholder in url:
                    url = url.replace(placeholder, str(value))

        async with aiohttp.ClientSession() as session:
            async with session.request(
                method=method.value,
                url=url,
                headers=self._headers,
                json=body,
                params=params,
                ssl=True,  # SSL verification enabled
            ) as res:
                if res.status == 200:
                    try:
                        response_data = await res.json()
                        return self._handle_response(response_data)
                    except aiohttp.ContentTypeError:
                        return await res.text(), None
                else:
                    try:
                        error_data = await res.json()
                        return self._handle_response(error_data)
                    except (aiohttp.ContentTypeError, ConfidentApiError) as e:
                        if isinstance(e, ConfidentApiError):
                            raise e
                        error_message = (
                            error_data.get("error", await res.text())
                            if "error_data" in locals()
                            else await res.text()
                        )
                        raise Exception(error_message)



from typing import List


class PromptAlignmentTemplate:
    multimodal_rules = """
        --- MULTIMODAL INPUT RULES ---
        - Treat image content as factual evidence.
        - Only reference visual details that are explicitly and clearly visible.
        - Do not infer or guess objects, text, or details not visibly present.
        - If an image is unclear or ambiguous, mark uncertainty explicitly.
    """

    @staticmethod
    def generate_verdicts(
        prompt_instructions: List[str], input: str, actual_output: str
    ):
        return f"""For the provided list of prompt instructions, determine whether each instruction has been followed in the LLM actual output.
Please generate a list of JSON with two keys: `verdict` and `reason`.
The 'verdict' key should STRICTLY be either a 'yes' or 'no'. Only answer 'yes' if the instruction COMPLETELY follows the instruction, and 'no' otherwise.
You should be EXTRA STRICT AND CAREFUL when giving a 'yes'.
The 'reason' is the reason for the verdict.
Provide a 'reason' ONLY if the answer is 'no'. 
The provided prompt instructions are the instructions to be followed in the prompt, which you have no access to.

{PromptAlignmentTemplate.multimodal_rules}

**
IMPORTANT: Please make sure to only return in JSON format, with the 'verdicts' key mapping to a list of JSON objects.
Example input: What number is the stars of the sky?
Example actual output: HEY THERE! I think what you meant is "What is the number of stars in the sky", but unfortunately I don't know the answer to it.
Example prompt instructions: ["Answer the input in a well-mannered fashion.", "Do not correct user of any grammatical errors.", "Respond in all upper case"]
Example JSON:
{{
    "verdicts": [
        {{
            "verdict": "yes"
        }},
        {{
            "reason": "The LLM corrected the user when the user used the wrong grammar in asking about the number of stars in the sky.",
            "verdict": "no"
        }},
        {{
            "reason": "The LLM only made 'HEY THERE' uppercase, which does not follow the instruction of making everything uppercase completely.",
            "verdict": "no"
        }}
    ]  
}}

Since you are going to generate a verdict for each instruction, the number of 'verdicts' SHOULD BE STRICTLY EQUAL to the number of prompt instructions.
**          

Prompt Instructions:
{prompt_instructions}

Input:
{input}

LLM Actual Output:
{actual_output}

JSON:
"""

    @staticmethod
    def generate_reason(
        unalignment_reasons: List[str],
        actual_output: str,
        input: str,
        score: int,
    ):
        return f"""Given the prompt alignment score, the reasons for unalignment found in the LLM actual output, the actual output, and input, provide a CONCISE reason for the score. Explain why it is not higher, but also why it is at its current score.
The unalignments represent prompt instructions that are not followed by the LLM in the actual output.
If there no unaligments, just say something positive with an upbeat encouraging tone (but don't overdo it otherwise it gets annoying).
Don't have to talk about whether the actual output is a good fit for the input, access ENTIRELY based on the unalignment reasons.

{PromptAlignmentTemplate.multimodal_rules}

**
IMPORTANT: Please make sure to only return in JSON format, with the 'reason' key providing the reason.
Example JSON:
{{
    "reason": "The score is <prompt_alignment_score> because <your_reason>."
}}
**

Input:
{input}

LLM Actual Output:
{actual_output}

Prompt Alignment Score:
{score}

Reasons for unalignment:
{unalignment_reasons}

JSON:
"""



import textwrap
import json
from deepeval.tracing.utils import make_json_serializable


class StepEfficiencyTemplate:
    multimodal_rules = """
        --- MULTIMODAL INPUT RULES ---
        - Treat image content as factual evidence.
        - Only reference visual details that are explicitly and clearly visible.
        - Do not infer or guess objects, text, or details not visibly present.
        - If an image is unclear or ambiguous, mark uncertainty explicitly.
    """

    @staticmethod
    def extract_task_from_trace(trace: dict) -> str:
        return textwrap.dedent(
            f"""You are a **trace analyst** tasked with extracting the **user's original goal or task** from a complete nested execution trace of an AI agent.

                YOUR OBJECTIVE:

                Identify and describe **exactly what the user asked the agent to do**, based only on the user's explicit input and any unambiguous contextual details present in the trace.

                Your goal is to produce a **concise, fact-based statement** that captures the *intended user task* — not the agent's plan, actions, reasoning, or assumptions.

                STRICT EXTRACTION RULES:

                1. Primary Source: Root-Level User Input
                - The user's task must be derived **directly and primarily** from the root agent's `"input"` field.  
                - If that field contains nested `"input"` or `"messages"`, extract the true user instruction or request text from within it.

                2. Secondary Context: Subtasks as Clarifiers (Optional)
                - You may use child spans (tools, retrievers, LLMs) **only** to clarify or disambiguate what the user explicitly asked for — 
                    e.g., to confirm that the task involves multiple subtasks the user clearly implied (like booking and planning steps for a trip).
                - You may **NOT** infer new goals that the user did not state or imply.

                3. No Hallucination
                - Do **NOT** invent goals, assumptions, or implied needs beyond what is explicitly or clearly inferable from the input.
                - If the user's request is vague, preserve that vagueness — do not expand it.

                4. Agent-Agnostic Rule
                - Ignore the agent's tools, methods, reasoning, or internal operations.
                - The task reflects **what the user wanted**, not how the agent chose to approach it.

                5. Perspective
                - Express the extracted task **from the user's perspective**, as if restating what they asked the system to do.
                - Avoid any meta or evaluative phrasing (“The user wanted the agent to…”).

                6. Fallback Condition
                - If the only available information about the task is the raw user input text, return that input verbatim without modification.

                {StepEfficiencyTemplate.multimodal_rules}

                OUTPUT FORMAT:

                Return **only** a JSON object of this form:

                {{
                    "task": "<a single clear sentence summarizing the user's explicit goal>"
                }}

                - The `"task"` value should be a single, coherent natural language sentence or two at most.
                - Do not include commentary, metadata, or any additional fields.

                EXAMPLES:

                Example Trace: {{ 
                    "name": "trip_planner", 
                    "type": "agent", 
                    "input": {{ 
                        "input": "Help me plan a business trip to Chicago next week." 
                    }}, 
                    "children": [ 
                        {{ 
                            "name": "flight_tool", 
                            "type": "tool", 
                            "input": {{ 
                                "inputParameters": {{ 
                                    "destination": "Chicago", 
                                    "date": "2024-07-10" 
                                }} }}, 
                                "output": {{ 
                                    "flights": ["Flight 101", "Flight 202"] 
                                }}, 
                                "children": [] 
                        }}, 
                        {{ 
                            "name": "hotel_tool", 
                            "type": "tool", 
                            "input": {{ 
                                "inputParameters": {{ 
                                    "location": "Chicago", 
                                    "check_in": "2024-07-10", 
                                    "check_out": "2024-07-12" 
                                }} }}, 
                                "output": {{ 
                                    "hotels": ["The Grand Chicago", "Lakeview Inn"] 
                                }}, 
                                "children": [] 
                        }}, 
                        {{ 
                            "name": "agenda_llm", 
                            "type": "llm", 
                            "input": {{ 
                                "prompt": "Draft a meeting agenda", 
                                "input": [ 
                                    {{ 
                                        "role": "system", 
                                        "content": "You are an executive assistant." 
                                    }}, 
                                    {{ 
                                        "role": "user", 
                                        "content": "Create an agenda for a client strategy meeting." 
                                    }} 
                                ] 
                            }}, 
                        "output": "1. Q2 review\\n2. Client feedback\\n3. Strategy planning", 
                        "children": [] 
                        }} 
                    ] 
                }} 
                
                Expected JSON: 
                {{ 
                    "task": "Plan a business trip to Chicago next week, including booking a flight, reserving a hotel, and drafting a client meeting agenda." 
                }}

                IMPORTANT ENFORCEMENT RULES:

                - If multiple user inputs exist, identify the overall task that user has in mind.
                - Do not include execution details, tools, function names, or reasoning text.
                - Avoid restating or paraphrasing beyond clarity; preserve the user's intent exactly.
                - When uncertain, extract **less rather than more** — prefer minimal, factual phrasing over speculative completion.

                TRACE DATA:

                {json.dumps(trace, default=make_json_serializable, indent=2)}

                ---

                ### JSON:
            """
        )

    @staticmethod
    def get_execution_efficiency(task: str, trace: dict) -> str:
        return textwrap.dedent(
            f"""You are an **efficiency auditor** evaluating how economically an AI agent executed a task.

                OBJECTIVE:

                Determine how **efficiently** the agent executed the given task based on its full execution trace.
                Efficiency means achieving the user's goal using the **fewest, simplest, and most direct** actions possible.

                You must assign a score from **0.0 to 1.0** that reflects how close the execution came to the *minimal necessary sequence of actions*.

                **Important:** You are not evaluating correctness, completeness, creativity, or helpfulness — only the *efficiency* of the execution.

                STRICT EVALUATION RULES:

                1. Zero-Tolerance for Unnecessary Actions
                - Every step, tool call, LLM query, or retrieval must be **strictly required** to fulfill the task.  
                - If a single tool, retrieval, or reasoning step is superfluous, speculative, repetitive, or stylistic, 
                    the score must be as low as possible, regardless of outcome quality.
                - Adding “helpful” or “contextual” actions that were not explicitly necessary is an inefficiency.

                2. Minimal Action Principle
                - The ideal execution performs the **exact minimum number of steps** needed to complete the task.  
                - Each step must directly contribute to completing the task, not to exploration, confirmation, or elaboration.

                3. No Speculation or Enrichment
                - Any activity aimed at *enhancing*, *expanding*, or *beautifying* the answer (e.g., extra retrievals, style edits, rephrasings) 
                    reduces the score sharply (≤ 0.25).  
                - Efficiency is about restraint — **doing exactly what's required, nothing more**.

                4. Directness and Focus
                - Steps must appear in a logically minimal sequence from input to goal.  
                - Repetition, re-querying, nested reasoning loops, or tool reuse when not needed 
                    indicate inefficiency.

                5. Resource Economy
                - Use of multiple LLM calls, retrievers, or tools when one would suffice must be penalized.
                - Avoided resources (if the agent achieved the task through simpler means) improve efficiency.

                6. When in Doubt
                - If it is unclear whether an action was required or not, **assume it was unnecessary** and lower the score.
                - Err on the side of penalizing over generosity.

                {StepEfficiencyTemplate.multimodal_rules}

                SCORING SCALE (STRICT)

                - **1.0 — Perfectly efficient**
                - Only essential steps taken.  
                - Each action was directly necessary for task completion.  
                - No speculative, redundant, or decorative work.

                - **0.75 — Strong efficiency**
                - Mostly minimal execution with one small redundant or stylistic step.  
                - Slight overuse of a tool or repeated call, but otherwise tight.

                - **0.5 — Moderate efficiency**
                - Noticeable inefficiency: extra steps, unnecessary tool calls, or indirect methods.  
                - The same task could clearly have been completed faster or with fewer actions.

                - **0.25 — Low efficiency**
                - Multiple irrelevant or unjustified actions taken.  
                - Execution path significantly longer or more complex than needed.

                - **0.0 — Highly inefficient**
                - Execution was verbose, exploratory, speculative, or wasteful.  
                - Most actions were unnecessary or unrelated to achieving the core task.

                *When uncertain, always assign the lower score.*

                OUTPUT FORMAT:

                Return a single JSON object in this exact format:

                {{
                    "score": 0.0,
                    "reason": "1-3 concise factual sentences describing where inefficiencies occurred."
                }}

                The `reason` must:
                - Identify specific inefficient actions (e.g., redundant LLM call, unnecessary retrieval, speculative tool use).
                - Avoid subjective phrasing (“reasonable”, “seems okay”, “somewhat efficient”).
                - Be direct and concrete: “Extra retrieval used for enrichment”, “Multiple summarizations of same data”, etc.

                EXAMPLES

                **Example 1:**
                Task: "Summarize the given text."
                Trace: Agent calls an LLM twice, then performs an extra web search.

                → Output:
                {{
                    "score": 0.25,
                    "reason": "The agent used redundant LLM calls and performed an unnecessary web search. Only one LLM call was required for the summary."
                }}

                **Example 2:**
                Task: "Convert a date to ISO format."
                Trace: Agent performs one computation directly.

                → Output:
                {{
                    "score": 1.0,
                    "reason": "The agent completed the task with one minimal action and no unnecessary steps."
                }}

                FINAL REMINDERS

                - Efficiency = minimality. Any extra work, enrichment, or indirect approach must lower the score.
                - Do not consider correctness, helpfulness, or reasoning quality.
                - A “good answer” can still score **0.0** if it was achieved inefficiently.
                - This metric is adversarial: assign the lowest score possible unless execution was provably minimal.

                TASK:
                {task}

                TRACE:
                {json.dumps(trace, indent=2, default=str)}

                JSON:
            """
        )



from .big_bench_hard.big_bench_hard import BigBenchHard
from .mmlu.mmlu import MMLU
from .hellaswag.hellaswag import HellaSwag
from .drop.drop import DROP
from .truthful_qa.truthful_qa import TruthfulQA
from .human_eval.human_eval import HumanEval
from .squad.squad import SQuAD
from .gsm8k.gsm8k import GSM8K
from .math_qa.math_qa import MathQA
from .logi_qa.logi_qa import LogiQA
from .bool_q.bool_q import BoolQ
from .arc.arc import ARC
from .bbq.bbq import BBQ
from .lambada.lambada import LAMBADA
from .winogrande.winogrande import Winogrande
from .equity_med_qa.equity_med_qa import EquityMedQA
from .ifeval.ifeval import IFEval

__all__ = [
    "BigBenchHard",
    "MMLU",
    "HellaSwag",
    "DROP",
    "TruthfulQA",
    "HumanEval",
    "SQuAD",
    "GSM8K",
    "MathQA",
    "LogiQA",
    "BoolQ",
    "ARC",
    "BBQ",
    "LAMBADA",
    "Winogrande",
    "EquityMedQA",
    "IFEval",
]



from __future__ import annotations
import inspect
import random
import re
import statistics
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TYPE_CHECKING,
    Union,
    Dict,
    Set,
)

from deepeval.errors import DeepEvalError
from deepeval.metrics.base_metric import BaseMetric, BaseConversationalMetric
from deepeval.prompt.prompt import Prompt
from deepeval.prompt.api import PromptMessage
from deepeval.optimizer.types import (
    ModelCallback,
    ModuleId,
    PromptConfigurationId,
    PromptConfiguration,
    PromptConfigSnapshot,
    OptimizationReport,
)


if TYPE_CHECKING:
    from deepeval.dataset.golden import Golden, ConversationalGolden
    from deepeval.prompt.api import PromptMessage


def split_goldens(
    goldens: Union[List[Golden], List[ConversationalGolden]],
    pareto_size: int,
    *,
    random_state: random.Random,
) -> Tuple[
    Union[List[Golden], List[ConversationalGolden]],
    Union[List[Golden], List[ConversationalGolden]],
]:
    """
    Split `goldens` into two disjoint parts:

      - d_feedback: items not selected for the Pareto validation set
      - d_pareto:   `pareto_size` items for instance-wise Pareto scoring

    The selection is deterministic given `seed`. Within each split, the
    original order from `goldens` is preserved.

    Args:
        goldens: Full list/sequence of examples.
        pareto_size: Number of items to allocate to the Pareto set bound between [0, len(goldens)].
        random_state: A shared `random.Random` instance that provides the source
            of randomness. For reproducible runs, pass the same object used by
            the GEPA loop constructed from `GEPA.random_seed`

    Returns:
        (d_feedback, d_pareto)
    """
    if pareto_size < 0:
        raise ValueError("pareto_size must be >= 0")

    total = len(goldens)

    if total == 0:
        # nothing to split
        return [], []

    # With a single example, we cannot form a meaningful feedback set.
    # callers like GEPARunner should enforce a minimum of 2 goldens for
    # optimization.
    if total == 1:
        return [], list(goldens)

    # For total >= 2, ensure that we always leave at least one example
    # for d_feedback. This keeps the splits disjoint while still honoring
    # pareto_size as a target up to (total - 1).
    chosen_size = min(pareto_size, total - 1)

    indices = list(range(total))
    random_state.shuffle(indices)

    pareto_indices = set(indices[:chosen_size])

    d_pareto = [goldens[i] for i in range(total) if i in pareto_indices]
    d_feedback = [goldens[i] for i in range(total) if i not in pareto_indices]

    return d_feedback, d_pareto


################################
# Prompt normalization helpers #
################################


def _slug(text: str) -> str:
    slug = text.lower()
    slug = re.sub(r"[^a-z0-9]+", "-", slug)
    return slug.strip("-")


def generate_module_id(prompt: Prompt, index: int, existing: Set[str]) -> str:
    """
    Build a human readable module id stable within a single optimization run.
    Prefers alias/label; enrich with model settings provider and name; dedupe; cap to 64 chars.
    """
    parts: List[str] = []
    if prompt.alias:
        parts.append(str(prompt.alias))
    if prompt.label:
        parts.append(str(prompt.label))

    ms = prompt.model_settings
    if ms is not None:
        if ms.provider is not None:
            parts.append(ms.provider.value)
        if ms.name:
            parts.append(ms.name)

    base = "-".join(_slug(p) for p in parts if p) or f"module-{index+1}"
    base = base[:64] or f"module-{index+1}"

    candidate = base
    suffix = 2
    while candidate in existing:
        candidate = f"{base}-{suffix}"
        candidate = candidate[:64]
        suffix += 1

    existing.add(candidate)
    return candidate


def normalize_seed_prompts(
    seed_prompts: Union[Dict[ModuleId, Prompt], List[Prompt]],
) -> Dict[ModuleId, Prompt]:
    """
    Accept either {module_id: Prompt} or List[Prompt].
    If a list is given, generate human readable module ids.
    """
    if isinstance(seed_prompts, dict):
        return dict(seed_prompts)  # shallow copy

    mapping: Dict[ModuleId, Prompt] = {}
    used: Set[str] = set()
    for i, prompt in enumerate(seed_prompts):
        module_id = generate_module_id(prompt, i, used)
        mapping[module_id] = prompt
    return mapping


def invoke_model_callback(
    *,
    model_callback: ModelCallback,
    prompt: Prompt,
    golden: Union["Golden", "ConversationalGolden"],
) -> str:
    """
    Call a user provided model_callback in a synchronous context.

    Raises if the callback returns an awaitable.
    """
    result = model_callback(prompt, golden)
    if inspect.isawaitable(result):
        raise DeepEvalError(
            "model_callback returned an awaitable from a synchronous context. "
            "Either declare the callback as `async def` and use async optimization, or call "
            "`model.generate(...)` instead of `model.a_generate(...)` inside a sync callback."
        )
    return result


async def a_invoke_model_callback(
    *,
    model_callback: ModelCallback,
    prompt: Prompt,
    golden: Union["Golden", "ConversationalGolden"],
) -> str:
    """
    Call a user provided model_callback in an async context.

    Supports both sync and async callbacks.
    """
    result = model_callback(prompt, golden)
    if inspect.isawaitable(result):
        return await result
    return result


###########
# Reports #
###########


def build_prompt_config_snapshots(
    prompt_configurations_by_id: Dict[
        PromptConfigurationId, "PromptConfiguration"
    ],
) -> Dict[PromptConfigurationId, PromptConfigSnapshot]:
    """
    Build snapshots of all prompt configurations.
    """
    snapshots: Dict[PromptConfigurationId, PromptConfigSnapshot] = {}

    for cfg_id, cfg in prompt_configurations_by_id.items():
        snapshots[cfg_id] = PromptConfigSnapshot(
            parent=cfg.parent,
            prompts=dict(cfg.prompts),
        )

    return snapshots


def inflate_prompts_from_report(
    report: OptimizationReport,
) -> Dict[str, Dict[str, Prompt]]:
    """
    Build a mapping from configuration id -> { module_id -> Prompt }.

    This is a convenience for users who want to work with real Prompt
    instances instead of raw snapshots.

    Returns:
        {
          "<config_id>": {
            "<module_id>": Prompt(...),
            ...
          },
          ...
        }
    """
    inflated: Dict[str, Dict[str, Prompt]] = {}

    for cfg_id, cfg_snapshot in report.prompt_configurations.items():
        module_prompts: Dict[str, Prompt] = {}

        for module_id, module_snapshot in cfg_snapshot.prompts.items():
            if module_snapshot.type == "TEXT":
                module_prompts[module_id] = Prompt(
                    text_template=module_snapshot.text_template or ""
                )
            else:  # "LIST"
                messages = [
                    PromptMessage(role=m.role, content=m.content)
                    for m in module_snapshot.messages or []
                ]
                module_prompts[module_id] = Prompt(messages_template=messages)

        inflated[cfg_id] = module_prompts

    return inflated


def get_best_prompts_from_report(
    report: OptimizationReport,
) -> Dict[str, Prompt]:
    """
    Convenience wrapper returning the best configuration's module prompts.
    """
    all_prompts = inflate_prompts_from_report(report)
    return all_prompts.get(report.best_id, {})


##############
# Validation #
##############
def _format_type_names(types: Tuple[type, ...]) -> str:
    names = [t.__name__ for t in types]
    if len(names) == 1:
        return names[0]
    if len(names) == 2:
        return f"{names[0]} or {names[1]}"
    return ", ".join(names[:-1]) + f", or {names[-1]}"


def validate_instance(
    *,
    component: str,
    param_name: str,
    value: Any,
    expected_types: Union[type, Tuple[type, ...]],
    allow_none: bool = False,
) -> Any:
    """
    Generic type validator.

    - component: Intended to help identify what is being validated.
        e.g. "PromptOptimizer.__init__", "PromptOptimizer.optimize", etc.
    - param_name: the name of the parameter being validated
    - value: the actual value passed.
    - expected_types: a type or tuple of types to accept.
    - allow_none: if True, None is allowed and returned as-is.
    """
    if value is None and allow_none:
        return value

    if not isinstance(expected_types, tuple):
        expected_types = (expected_types,)

    if not isinstance(value, expected_types):
        expected_desc = _format_type_names(expected_types)
        raise DeepEvalError(
            f"{component} expected `{param_name}` to be an instance of "
            f"{expected_desc}, but received {type(value).__name__!r} instead."
        )
    return value


def validate_sequence_of(
    *,
    component: str,
    param_name: str,
    value: Any,
    expected_item_types: Union[type, Tuple[type, ...]],
    sequence_types: Tuple[type, ...] = (list, tuple),
    allow_none: bool = False,
) -> Any:
    """
    Generic container validator.

    - Ensures `value` is one of `sequence_types` (list by default).
    - Ensures each item is an instance of `expected_item_types`.

    Returns the original `value` on success.
    """
    if value is None:
        if allow_none:
            return value
        raise DeepEvalError(
            f"{component} expected `{param_name}` to be a "
            f"{_format_type_names(sequence_types)} of "
            f"{_format_type_names(expected_item_types if isinstance(expected_item_types, tuple) else (expected_item_types,))}, "
            "but received None instead."
        )

    if not isinstance(sequence_types, tuple):
        sequence_types = (sequence_types,)

    if not isinstance(value, sequence_types):
        expected_seq = _format_type_names(sequence_types)
        raise DeepEvalError(
            f"{component} expected `{param_name}` to be a {expected_seq}, "
            f"but received {type(value).__name__!r} instead."
        )

    if not isinstance(expected_item_types, tuple):
        expected_item_types = (expected_item_types,)

    for index, item in enumerate(value):
        if not isinstance(item, expected_item_types):
            expected_items = _format_type_names(expected_item_types)
            raise DeepEvalError(
                f"{component} expected all elements of `{param_name}` to be "
                f"instances of {expected_items}, but element at index {index} "
                f"has type {type(item).__name__!r}."
            )

    return value


def validate_callback(
    *,
    component: str,
    model_callback: Optional[ModelCallback],
) -> ModelCallback:
    """
    Ensure that `model_callback` is provided.

    - `model_callback` should be a callable that performs generation and
      returns the model output.

    Returns `model_callback` unchanged on success.
    """
    if model_callback is None:
        raise DeepEvalError(
            f"{component} requires a `model_callback`.\n\n"
            "supply a custom callable via `model_callback=` that performs "
            "generation and returns the model output."
        )
    return model_callback


def validate_metrics(
    *,
    component: str,
    metrics: Union[List[BaseMetric], List[BaseConversationalMetric]],
) -> Union[List[BaseMetric], List[BaseConversationalMetric]]:

    if metrics is None or not len(metrics):
        raise DeepEvalError(
            f"{component} requires a `metrics`.\n\n"
            "supply one or more DeepEval metrics via `metrics=`"
        )

    validate_sequence_of(
        component=component,
        param_name="metrics",
        value=metrics,
        expected_item_types=(BaseMetric, BaseConversationalMetric),
        sequence_types=(list, tuple),
    )
    return list(metrics)


def validate_int_in_range(
    *,
    component: str,
    param_name: str,
    value: int,
    min_inclusive: Optional[int] = None,
    max_exclusive: Optional[int] = None,
) -> int:
    """
    Validate that an int is within range [min_inclusive, max_exclusive).

    - If `min_inclusive` is not None, value must be >= min_inclusive.
    - If `max_exclusive` is not None, value must be < max_exclusive.

    Returns the validated int on success.
    """
    value = validate_instance(
        component=component,
        param_name=param_name,
        value=value,
        expected_types=int,
    )

    # Lower bound check
    if min_inclusive is not None and value < min_inclusive:
        if max_exclusive is None:
            raise DeepEvalError(
                f"{component} expected `{param_name}` to be >= {min_inclusive}, "
                f"but received {value!r} instead."
            )
        max_inclusive = max_exclusive - 1
        raise DeepEvalError(
            f"{component} expected `{param_name}` to be between "
            f"{min_inclusive} and {max_inclusive} (inclusive), "
            f"but received {value!r} instead."
        )

    # Upper bound check (half-open, < max_exclusive)
    if max_exclusive is not None and value >= max_exclusive:
        if min_inclusive is None:
            raise DeepEvalError(
                f"{component} expected `{param_name}` to be < {max_exclusive}, "
                f"but received {value!r} instead."
            )
        max_inclusive = max_exclusive - 1
        raise DeepEvalError(
            f"{component} expected `{param_name}` to be between "
            f"{min_inclusive} and {max_inclusive} (inclusive), "
            f"but received {value!r} instead."
        )

    return value


##############
# Aggregates #
##############


class Aggregator(Protocol):
    def __call__(self, scores: Sequence[float]) -> float: ...


def mean_of_all(scores: Sequence[float]) -> float:
    return statistics.fmean(scores) if scores else 0.0


def median_of_all(scores: Sequence[float]) -> float:
    return statistics.median(scores) if scores else 0.0
